# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from typing import Union


class TapqirException(Exception):
    """
    Base class for all tapqir exception.
    """

    def __init__(self, msg, *args):
        assert msg
        self.msg = msg
        super().__init__(msg, *args)


class TapqirFileNotFoundError(TapqirException):
    """
    Thrown if a file/directory is not found as an output in any pipeline.

    :param name: file type name.
    :param path: file path.
    """

    def __init__(self, name: str, path: Union[str, Path]):
        self.name = name
        self.path = path
        super().__init__(f"Unable to find {name} file '{path}'")


class CudaOutOfMemoryError(TapqirException):
    """
    Thrown if CUDA is out of memory.
    """

    def __init__(self):
        super().__init__("CUDA out of memory. Try to use smaller AOI/frame batch size")



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

"""
KSMOGN
^^^^^^
"""

from typing import Union

import pykeops
import torch
from pykeops.torch import Genred
from pyro.distributions import TorchDistribution
from torch.distributions import Categorical, constraints
from torch.distributions.utils import lazy_property

from .util import gaussian_spots

pykeops.set_verbose(False)


class KSMOGN(TorchDistribution):
    r"""
    K-Spots Marginalized Offset Gamma Noise Image Distribution.

    .. math::
        \mu^S_{\mathsf{pixelX}(i), \mathsf{pixelY}(j)} =
        \dfrac{m \cdot h}{2 \pi w^2}
        \exp{\left( -\dfrac{(i-x-x^\mathsf{target})^2 + (j-y-y^\mathsf{target})^2}{2 w^2} \right)}

    .. math::
        \mu^I = b + \sum_{\mathsf{spot}} \mu^S

    .. math::
        p(D|\mu^I, g) = \sum_\delta p(\delta) p(D|\mu^I, g, \delta) = \sum_\delta \delta_\mathsf{weights}
        \cdot \mathrm{Gamma}(D - \delta_\mathsf{samples} | \mu^I, g)

    **Reference**:

    1. Ordabayev YA, Friedman LJ, Gelles J, Theobald DL.
       Bayesian machine learning analysis of single-molecule fluorescence colocalization images.
       bioRxiv. 2021 Oct. doi: `10.1101/2021.09.30.462536 <https://doi.org/10.1101/2021.09.30.462536>`_.

    :param height: Integrated spot intensity. Should be broadcastable
        to ``batch_shape + (K,)``.
    :param width: Spot width. Should be broadcastable
        to ``batch_shape + (K,)``.
    :param x: Spot center on x-axis. Should be broadcastable
        to ``batch_shape + (K,)``.
    :param y: Spot center on y-axis. Should be broadcastable
        to ``batch_shape + (K,)``.
    :param target_locs: Target location. Should have
        the rightmost size ``2`` correspondnig to locations on
        x- and y-axes, and be broadcastable to ``batch_shape + (2,)``.
    :param background: Background intensity. Should
        be broadcastable to ``batch_shape``.
    :param gain: Camera gain.
    :param offset_samples: Offset samples from the empirical distribution.
    :param offset_logits: Offset log weights corresponding to the offset samples.
    :param int P: Number of pixels along the axis.
    :param m: Spot presence indicator. Should be broadcastable
        to ``batch_shape + (K,)``.
    :param alpha: Signal cross-talk coefficient matrix. Should be broadcastable
        to ``(Q, C)``.
    :param bool use_pykeops: Use pykeops as backend to marginalize out offset.
    """

    arg_constraints = {}
    support = constraints.positive

    def __init__(
        self,
        height: torch.Tensor,
        width: torch.Tensor,
        x: torch.Tensor,
        y: torch.Tensor,
        target_locs: torch.Tensor,
        background: torch.Tensor,
        gain: Union[float, torch.Tensor],
        offset_samples: torch.Tensor,
        offset_logits: torch.Tensor,
        P: int,
        m: torch.Tensor = None,
        alpha: torch.Tensor = None,
        use_pykeops: bool = True,
        validate_args=None,
    ):

        self.height = height
        self.width = width
        self.x = x
        self.y = y
        self.target_locs = target_locs
        self.m = m
        self.background = background[..., None, None]
        if alpha is not None:
            C = alpha.shape[-1]
            self.gain = gain[..., None, None, None]
            self.alpha = alpha[..., None, None, None]
        else:
            self.gain = gain[..., None, None]
            self.alpha = alpha
        self.rate = 1 / self.gain

        self.offset_samples = offset_samples
        self.offset_logits = offset_logits
        self.P = P
        self.use_pykeops = use_pykeops
        if self.use_pykeops:
            device = self.target_locs.device.type
            self.device_pykeops = "GPU" if device == "cuda" else "CPU"

        # calculate batch shape
        batch_shape = torch.broadcast_shapes(
            height.shape, width.shape, x.shape, y.shape
        )
        if m is not None:
            batch_shape = torch.broadcast_shapes(batch_shape, m.shape)

        event_shape = torch.Size([P, P])
        bg_shape = background.shape
        target_shape = target_locs.shape[:-1]
        # remove K dim
        batch_shape = batch_shape[:-1]
        if alpha is not None:
            # remove Q dim
            batch_shape = batch_shape[:-1]
            # add C dim
            event_shape = (C,) + event_shape
            # remove C dim
            bg_shape = bg_shape[:-1]
            target_shape = target_shape[:-1]
        batch_shape = torch.broadcast_shapes(batch_shape, bg_shape, target_shape)
        super().__init__(batch_shape, event_shape, validate_args=validate_args)

    @lazy_property
    def gaussians(self):
        return gaussian_spots(
            self.height,
            self.width,
            self.x,
            self.y,
            self.target_locs.unsqueeze(-2),
            self.P,
            self.m,
        )

    @lazy_property
    def image(self):
        gaussians = self.gaussians
        if self.alpha is not None:
            # sum over the signal from multiple fluorescent dyes
            gaussians = (gaussians.unsqueeze(-4) * self.alpha).sum(-5)
        # sum the background and all the spots
        return self.background + gaussians.sum(-3)

    @lazy_property
    def concentration(self):
        return self.image / self.gain

    def rsample(self, sample_shape=torch.Size()):
        odx = (
            Categorical(logits=self.offset_logits)
            .expand(self.batch_shape + (self.P, self.P))
            .sample()
        )
        offset = self.offset_samples[odx]
        shape = self._extended_shape(sample_shape)
        value = torch._standard_gamma(
            self.concentration.expand(shape)
        ) / self.rate.expand(shape)
        value.detach().clamp_(
            min=torch.finfo(value.dtype).tiny
        )  # do not record in autograd graph
        return value + offset

    def log_prob(self, value):
        if self.use_pykeops:
            formula = "wj+Log(Step(xi-gj-IntCst(1)))+(ai-IntCst(1))*Log(IfElse(xi-gj-IntCst(1),xi-gj,xi))-bi*(xi-gj)"
            variables = [
                "wj = Vj(1)",
                "gj = Vj(1)",
                "ai = Vi(1)",
                "bi = Vi(1)",
                "xi = Vi(1)",
            ]
            dtype = self.concentration.dtype
            my_routine = Genred(
                formula,
                variables,
                reduction_op="LogSumExp",
                axis=1,
            )
            concentration, value, rate = torch.broadcast_tensors(
                self.concentration, value, self.rate
            )
            shape = value.shape
            result = my_routine(
                self.offset_logits.reshape(-1, 1),
                self.offset_samples.reshape(-1, 1).to(dtype),
                concentration.reshape(-1, 1),
                rate.reshape(-1, 1).contiguous(),
                value.reshape(-1, 1).to(dtype),
                backend=self.device_pykeops,
            )
            result = result.reshape(shape)
            result = (
                self.concentration * torch.log(self.rate)
                - torch.lgamma(self.concentration)
                + result
            )
        else:
            value = torch.as_tensor(value).unsqueeze(-1)
            concentration = self.concentration.unsqueeze(-1)
            mask = value > self.offset_samples
            new_value = torch.where(
                mask, value - self.offset_samples, value.new_ones(())
            )
            obs_logits = (
                concentration * torch.log(self.rate)
                + (concentration - 1) * torch.log(new_value)
                - self.rate * (new_value)
                - torch.lgamma(concentration)
            )
            result = obs_logits + self.offset_logits + torch.log(mask)
            result = torch.logsumexp(result, -1)
        event_dims = tuple(-i for i in range(1, len(self.event_shape) + 1))
        return result.sum(event_dims)



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import torch
from pyro.distributions import AffineBeta as PyroAffineBeta
from pyro.distributions.util import broadcast_shape
from torch.distributions import constraints


class AffineBeta(PyroAffineBeta):
    r"""
    Beta distribution shifted by :attr:`loc` and scaled by :attr:`scale`::

        concentration1 = size * (mean - low) / (high - low)
        concentration0 = size * (high - mean) / (high - low)
        X ~ Beta(concentration1, concentration0)
        f(X) = low + (high - low) * X
        Y = f(X) ~ AffineBeta(mean, sample_size, low, high)

    :param mean: mean of the distribution.
    :param sample_size: sample size parameter of the Beta distribution.
    :param low: min parameter.
    :param high: max parameter.
    """

    arg_constraints = {
        "mean": constraints.dependent,
        "sample_size": constraints.real,
        "low": constraints.real,
        "high": constraints.dependent,
    }

    def __init__(self, mean, sample_size, low, high, validate_args=None):
        if low != high:
            concentration1 = sample_size * (mean - low) / (high - low)
            concentration0 = sample_size * (high - mean) / (high - low)
        else:
            # this is needed to work with funsor make_dist
            low = torch.tensor(0.0)
            high = torch.tensor(1.0)
            concentration1 = torch.tensor(1.0)
            concentration0 = torch.tensor(1.0)
        super(AffineBeta, self).__init__(
            concentration1,
            concentration0,
            low,
            high - low,
            validate_args=validate_args,
        )

    @staticmethod
    def infer_shapes(mean, sample_size, low, high):
        batch_shape = broadcast_shape(mean, sample_size, low, high)
        event_shape = torch.Size()
        return batch_shape, event_shape

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(AffineBeta, _instance)
        return super(AffineBeta, self).expand(batch_shape, _instance=new)



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import re

import scipy.stats as stats
import torch
from pyro.poutine.messenger import Messenger
from pyroapi import distributions as dist


class StatsMessenger(Messenger):
    """
    Confidence interval with equal areas around the median.
    """

    def __init__(
        self,
        CI: float = 0.95,
        K: int = None,
        N: int = None,
        F: int = None,
        Q: int = None,
    ):
        super().__init__()
        self.CI = CI
        self.ci_stats = {}
        self.K = K
        self.N = N
        self.F = F
        self.Q = Q

    def _pyro_sample(self, msg):
        if (
            type(msg["fn"]).__name__ == "_Subsample"
            or msg["infer"].get("enumerate", None) == "parallel"
        ):
            return
        name = msg["name"]
        scipy_dist = torch_to_scipy_dist(msg["fn"])
        if scipy_dist is None:
            return
        LL, UL = scipy_dist.interval(alpha=self.CI)
        args = re.split("_k|_q", name)
        if len(args) == 1:
            (base_name,) = args
            self.ci_stats[base_name] = {}
            self.ci_stats[base_name]["LL"] = torch.as_tensor(
                LL, device=torch.device("cpu")
            )
            self.ci_stats[base_name]["UL"] = torch.as_tensor(
                UL, device=torch.device("cpu")
            )
            self.ci_stats[base_name]["Mean"] = msg["fn"].mean.detach().cpu()
        elif len(args) == 2:
            base_name, k = args
            k = int(k)
            assert self.Q == 1
            if k == 0:
                self.ci_stats[base_name] = {}
                self.ci_stats[base_name]["LL"] = torch.zeros(
                    self.K, self.N, self.F, device=torch.device("cpu")
                )
                self.ci_stats[base_name]["UL"] = torch.zeros(
                    self.K, self.N, self.F, device=torch.device("cpu")
                )
                self.ci_stats[base_name]["Mean"] = torch.zeros(
                    self.K, self.N, self.F, device=torch.device("cpu")
                )
            self.ci_stats[base_name]["LL"][k] = torch.as_tensor(
                LL, device=torch.device("cpu")
            )
            self.ci_stats[base_name]["UL"][k] = torch.as_tensor(
                UL, device=torch.device("cpu")
            )
            self.ci_stats[base_name]["Mean"][k] = msg["fn"].mean.detach().cpu()
        elif len(args) == 3:
            base_name, k, q = args
            k, q = int(k), int(q)
            assert self.Q > 1
            if (k == 0) and (q == 0):
                self.ci_stats[base_name] = {}
                self.ci_stats[base_name]["LL"] = torch.zeros(
                    self.K, self.N, self.F, self.Q, device=torch.device("cpu")
                )
                self.ci_stats[base_name]["UL"] = torch.zeros(
                    self.K, self.N, self.F, self.Q, device=torch.device("cpu")
                )
                self.ci_stats[base_name]["Mean"] = torch.zeros(
                    self.K, self.N, self.F, self.Q, device=torch.device("cpu")
                )
            self.ci_stats[base_name]["LL"][k, :, :, q] = torch.as_tensor(
                LL, device=torch.device("cpu")
            )
            self.ci_stats[base_name]["UL"][k, :, :, q] = torch.as_tensor(
                UL, device=torch.device("cpu")
            )
            self.ci_stats[base_name]["Mean"][k, :, :, q] = msg["fn"].mean.detach().cpu()
        msg["stop"] = True
        msg["done"] = True

    def __enter__(self):
        super().__enter__()
        return self.ci_stats


def torch_to_scipy_dist(torch_dist):
    if isinstance(torch_dist, dist.Gamma):
        return stats.gamma(
            torch_dist.concentration.detach().cpu(),
            scale=1 / torch_dist.rate.detach().cpu(),
        )
    elif isinstance(torch_dist, dist.Beta):
        return stats.beta(
            a=torch_dist.concentration1.detach().cpu(),
            b=torch_dist.concentration0.detach().cpu(),
        )
    elif isinstance(torch_dist, dist.AffineBeta):
        return stats.beta(
            a=torch_dist.concentration1.detach().cpu(),
            b=torch_dist.concentration0.detach().cpu(),
            loc=torch_dist.loc.detach().cpu(),
            scale=torch_dist.scale.detach().cpu(),
        )
    elif isinstance(torch_dist, dist.Dirichlet):
        return stats.beta(
            a=torch_dist.concentration.detach().cpu(),
            b=(
                torch_dist.concentration.sum(-1, keepdim=True).detach()
                - torch_dist.concentration.detach()
            ).cpu(),
        )
    elif isinstance(torch_dist, dist.Independent):
        return torch_to_scipy_dist(torch_dist.base_dist)
    elif isinstance(torch_dist, dist.Delta):
        return None
    else:
        raise NotImplementedError



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import logging
from collections import namedtuple
from pathlib import Path

import torch
from pyro.ops.indexing import Vindex
from pyro.ops.stats import quantile
from torch.distributions.utils import lazy_property, probs_to_logits

from tapqir.exceptions import TapqirFileNotFoundError

logger = logging.getLogger(__name__)


class OffsetData(namedtuple("OffsetData", ["samples", "weights"])):
    @lazy_property
    def min(self):
        return torch.min(self.samples).item()

    @lazy_property
    def max(self):
        return torch.max(self.samples).item()

    @lazy_property
    def logits(self):
        return probs_to_logits(self.weights)

    @lazy_property
    def mean(self):
        return torch.sum(self.samples * self.weights).item()

    @lazy_property
    def var(self):
        return torch.sum(self.samples**2 * self.weights).item() - self.mean**2


class CosmosDataset:
    """
    Cosmos Dataset
    """

    def __init__(
        self,
        images,
        xy,
        is_ontarget,
        mask=None,
        labels=None,
        offset_samples=None,
        offset_weights=None,
        device=torch.device("cpu"),
        time1=None,
        ttb=None,
        name=None,
    ):
        self.images = images
        self.xy = xy
        self.is_ontarget = is_ontarget
        if mask is None:
            mask = torch.ones_like(is_ontarget, dtype=torch.bool)
        self.mask = mask
        self.labels = labels
        self.device = device
        self.offset = OffsetData(offset_samples.to(device), offset_weights.to(device))
        self.time1 = time1
        self.ttb = ttb
        self.name = name

    @lazy_property
    def N(self) -> int:
        """
        Number of on-target AOIs.
        """
        return self.is_ontarget.sum().item()

    @lazy_property
    def Nc(self) -> int:
        """
        Number of off-target AOIs.
        """
        return (~self.is_ontarget).sum().item()

    @lazy_property
    def Nt(self) -> int:
        """
        Total number of AOIs.
        """
        return self.N + self.Nc

    @property
    def F(self) -> int:
        """
        Number of frames.
        """
        return self.images.shape[1]

    @property
    def C(self) -> int:
        """
        Number of color-channels.
        """
        return self.images.shape[2]

    @property
    def P(self) -> int:
        """
        Number of pixels.
        """
        return self.images.shape[3]

    @property
    def x(self) -> torch.Tensor:
        """
        Target location on the x-axis.
        """
        return self.xy[..., 0]

    @property
    def y(self) -> torch.Tensor:
        """
        Target location on the y-axis.
        """
        return self.xy[..., 1]

    @lazy_property
    def median(self) -> torch.Tensor:
        return torch.stack(
            [torch.median(self.images[..., c, :, :]) for c in range(self.C)]
        )

    def fetch(self, ndx, fdx, cdx):
        return (
            Vindex(self.images)[ndx, fdx, cdx].to(self.device),
            Vindex(self.xy)[ndx, fdx, cdx].to(self.device),
            Vindex(self.is_ontarget)[ndx].to(self.device),
        )

    @lazy_property
    def vmin(self) -> torch.Tensor:
        return torch.stack(
            [
                quantile(self.images[..., c, :, :].flatten().float(), 0.05)
                for c in range(self.C)
            ]
        )

    @lazy_property
    def vmax(self) -> int:
        return torch.stack(
            [
                quantile(self.images[..., c, :, :].flatten().float(), 0.99)
                for c in range(self.C)
            ]
        )

    def __repr__(self):
        samples = repr(self.offset.samples).replace("\n", "\n                  ")
        weights = repr(self.offset.weights).replace("\n", "\n                  ")
        return (
            f"{self.__class__.__name__}: {self.name}"
            f"\n  images           tensor(N={self.N} on-target AOIs, "
            f"Nc={self.Nc} off-target AOIs, "
            f"F={self.F} frames, "
            f"C={self.C} channels, "
            f"P={self.P} pixels, "
            f"P={self.P} pixels)"
            f"\n  x                tensor(N={self.N} on-target AOIs, "
            f"Nc={self.Nc} off-target AOIs, "
            f"F={self.F} frames, "
            f"C={self.C} channels)"
            f"\n  y                tensor(N={self.N} on-target AOIs, "
            f"Nc={self.Nc} off-target AOIs, "
            f"F={self.F} frames, "
            f"C={self.C} channels)"
            f"\n\n  offset.samples   {samples}"
            f"\n        .weights   {weights}"
        )


def save(obj, path):
    path = Path(path)
    torch.save(
        {
            "images": obj.images,
            "xy": obj.xy,
            "is_ontarget": obj.is_ontarget,
            "mask": obj.mask,
            "labels": obj.labels,
            "offset_samples": obj.offset.samples,
            "offset_weights": obj.offset.weights,
            "name": obj.name,
            "time1": obj.time1,
            "ttb": obj.ttb,
        },
        path / "data.tpqr",
    )
    logger.info(f"Data is saved in {path / 'data.tpqr'}")


def load(path, device=torch.device("cpu")):
    path = Path(path)
    try:
        data_tapqir = torch.load(path / "data.tpqr")
    except FileNotFoundError:
        raise TapqirFileNotFoundError("data", path / "data.tpqr")
    return CosmosDataset(**data_tapqir, **{"device": device})



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

"""
cosmos
^^^^^^
"""

import itertools
import math
from functools import reduce
from typing import Union

import torch
import torch.distributions.constraints as constraints
from pyro.ops.indexing import Vindex
from pyroapi import distributions as dist
from pyroapi import handlers, infer, pyro
from torch.distributions.utils import lazy_property
from torch.nn.functional import one_hot

from tapqir.distributions import KSMOGN, AffineBeta
from tapqir.distributions.util import expand_offtarget, probs_m, probs_theta
from tapqir.models.model import Model


class cosmos(Model):
    r"""
    **Single-Color Time-Independent Colocalization Model**

    **Reference**:

    1. Ordabayev YA, Friedman LJ, Gelles J, Theobald DL.
       Bayesian machine learning analysis of single-molecule fluorescence colocalization images.
       eLife. 2022 March. doi: `10.7554/eLife.73860 <https://doi.org/10.7554/eLife.73860>`_.

    :param S: Number of distinct molecular states for the binder molecules.
    :param K: Maximum number of spots that can be present in a single image.
    :param channels: Number of color channels.
    :param device: Computation device (cpu or gpu).
    :param dtype: Floating point precision.
    :param use_pykeops: Use pykeops as backend to marginalize out offset.
    """

    name = "cosmos"

    def __init__(
        self,
        K: int = 2,
        channels: Union[tuple, list] = (0,),
        device: str = "cpu",
        dtype: str = "double",
        use_pykeops: bool = True,
        priors: dict = {
            "background_mean_std": 1000,
            "background_std_std": 100,
            "lamda_rate": 1,
            "height_std": 10000,
            "width_min": 0.75,
            "width_max": 2.25,
            "proximity_rate": 1,
            "gain_std": 50,
        },
    ):
        S, Q = 1, 1
        super().__init__(S, K, Q, channels, device, dtype, priors)
        assert self.C == 1, "Please specify exactly one color channel"
        self._global_params = ["gain", "proximity", "lamda", "pi"]
        self.use_pykeops = use_pykeops
        self.conv_params = ["-ELBO", "proximity_loc", "gain_loc", "lamda_loc"]

    def model(self):
        r"""
        **Generative Model**

        Model parameters:

        +-----------------+-----------+-------------------------------------+
        | Parameter       | Shape     | Description                         |
        +=================+===========+=====================================+
        | |g| - :math:`g` | (1,)      | camera gain                         |
        +-----------------+-----------+-------------------------------------+
        | |sigma| - |prox|| (1,)      | proximity                           |
        +-----------------+-----------+-------------------------------------+
        | ``lamda`` - |ld|| (1,)      | average rate of target-nonspecific  |
        |                 |           | binding                             |
        +-----------------+-----------+-------------------------------------+
        | ``pi`` - |pi|   | (1,)      | average binding probability of      |
        |                 |           | target-specific binding             |
        +-----------------+-----------+-------------------------------------+
        | |bg| - |b|      | (N, F)    | background intensity                |
        +-----------------+-----------+-------------------------------------+
        | |z| - :math:`z` | (N, F)    | target-specific spot presence       |
        +-----------------+-----------+-------------------------------------+
        | |t| - |theta|   | (N, F)    | target-specific spot index          |
        +-----------------+-----------+-------------------------------------+
        | |m| - :math:`m` | (K, N, F) | spot presence indicator             |
        +-----------------+-----------+-------------------------------------+
        | |h| - :math:`h` | (K, N, F) | spot intensity                      |
        +-----------------+-----------+-------------------------------------+
        | |w| - :math:`w` | (K, N, F) | spot width                          |
        +-----------------+-----------+-------------------------------------+
        | |x| - :math:`x` | (K, N, F) | spot position on x-axis             |
        +-----------------+-----------+-------------------------------------+
        | |y| - :math:`y` | (K, N, F) | spot position on y-axis             |
        +-----------------+-----------+-------------------------------------+
        | |D| - :math:`D` | |shape|   | observed images                     |
        +-----------------+-----------+-------------------------------------+

        .. |ps| replace:: :math:`p(\mathsf{specific})`
        .. |theta| replace:: :math:`\theta`
        .. |prox| replace:: :math:`\sigma^{xy}`
        .. |ld| replace:: :math:`\lambda`
        .. |b| replace:: :math:`b`
        .. |shape| replace:: (N, F, P, P)
        .. |sigma| replace:: ``proximity``
        .. |bg| replace:: ``background``
        .. |h| replace:: ``height``
        .. |w| replace:: ``width``
        .. |D| replace:: ``data``
        .. |m| replace:: ``m``
        .. |z| replace:: ``z``
        .. |t| replace:: ``theta``
        .. |x| replace:: ``x``
        .. |y| replace:: ``y``
        .. |pi| replace:: :math:`\pi`
        .. |g| replace:: ``gain``

        Full joint distribution:

        .. math::

            \begin{aligned}
                p(D, \phi) =~&p(g) p(\sigma^{xy}) p(\pi) p(\lambda)
                \prod_{\mathsf{AOI}} \left[ p(\mu^b) p(\sigma^b) \prod_{\mathsf{frame}}
                \left[ \vphantom{\prod_{F}} p(b | \mu^b, \sigma^b) p(z | \pi) p(\theta | z)
                \vphantom{\prod_{\substack{\mathsf{pixelX} \\ \mathsf{pixelY}}}} \cdot \right. \right. \\
                &\prod_{\mathsf{spot}} \left[ \vphantom{\prod_{F}} p(m | \theta, \lambda)
                p(h) p(w) p(x | \sigma^{xy}, \theta) p(y | \sigma^{xy}, \theta) \right] \left. \left.
                \prod_{\substack{\mathsf{pixelX} \\ \mathsf{pixelY}}} \sum_{\delta} p(\delta)
                p(D | \mu^I, g, \delta) \right] \right]
            \end{aligned}

        :math:`z` and :math:`\theta` marginalized joint distribution:

        .. math::

            \begin{aligned}
                \sum_{z, \theta} p(D, \phi) =~&p(g) p(\sigma^{xy}) p(\pi) p(\lambda)
                \prod_{\mathsf{AOI}} \left[ p(\mu^b) p(\sigma^b) \prod_{\mathsf{frame}}
                \left[ \vphantom{\prod_{F}} p(b | \mu^b, \sigma^b) \sum_{z} p(z | \pi) \sum_{\theta} p(\theta | z)
                \vphantom{\prod_{\substack{\mathsf{pixelX} \\ \mathsf{pixelY}}}} \cdot \right. \right. \\
                &\prod_{\mathsf{spot}} \left[ \vphantom{\prod_{F}} p(m | \theta, \lambda)
                p(h) p(w) p(x | \sigma^{xy}, \theta) p(y | \sigma^{xy}, \theta) \right] \left. \left.
                \prod_{\substack{\mathsf{pixelX} \\ \mathsf{pixelY}}} \sum_{\delta} p(\delta)
                p(D | \mu^I, g, \delta) \right] \right]
            \end{aligned}
        """
        # global parameters
        gain = pyro.sample("gain", dist.HalfNormal(self.priors["gain_std"]))
        pi = pyro.sample("pi", dist.Dirichlet(torch.ones(self.S + 1) / (self.S + 1)))
        pi = expand_offtarget(pi)
        lamda = pyro.sample("lamda", dist.Exponential(self.priors["lamda_rate"]))
        proximity = pyro.sample(
            "proximity", dist.Exponential(self.priors["proximity_rate"])
        )
        size = torch.stack(
            (
                torch.full_like(proximity, 2.0),
                (((self.data.P + 1) / (2 * proximity)) ** 2 - 1),
            ),
            dim=-1,
        )

        # spots
        spots = pyro.plate("spots", self.K)
        # aoi sites
        aois = pyro.plate(
            "aois",
            self.data.Nt,
            subsample=self.n,
            subsample_size=self.nbatch_size,
            dim=-2,
        )
        # time frames
        frames = pyro.plate(
            "frames",
            self.data.F,
            subsample=self.f,
            subsample_size=self.fbatch_size,
            dim=-1,
        )

        with aois as ndx:
            ndx = ndx[:, None]
            mask = Vindex(self.data.mask)[ndx].to(self.device)
            with handlers.mask(mask=mask):
                # background mean and std
                background_mean = pyro.sample(
                    "background_mean",
                    dist.HalfNormal(self.priors["background_mean_std"]),
                )
                background_std = pyro.sample(
                    "background_std", dist.HalfNormal(self.priors["background_std_std"])
                )
                with frames as fdx:
                    # fetch data
                    obs, target_locs, is_ontarget = self.data.fetch(ndx, fdx, self.cdx)
                    # sample background intensity
                    background = pyro.sample(
                        "background",
                        dist.Gamma(
                            (background_mean / background_std) ** 2,
                            background_mean / background_std**2,
                        ),
                    )

                    # sample hidden model state (1+S,)
                    z = pyro.sample(
                        "z",
                        dist.Categorical(Vindex(pi)[..., :, is_ontarget.long()]),
                        infer={"enumerate": "parallel"},
                    )
                    theta = pyro.sample(
                        "theta",
                        dist.Categorical(
                            Vindex(probs_theta(self.K, self.device))[
                                torch.clamp(z, min=0, max=1)
                            ]
                        ),
                        infer={"enumerate": "parallel"},
                    )
                    onehot_theta = one_hot(theta, num_classes=1 + self.K)

                    ms, heights, widths, xs, ys = [], [], [], [], []
                    for kdx in spots:
                        specific = onehot_theta[..., 1 + kdx]
                        # spot presence
                        m = pyro.sample(
                            f"m_k{kdx}",
                            dist.Bernoulli(
                                Vindex(probs_m(lamda, self.K))[..., theta, kdx]
                            ),
                        )
                        with handlers.mask(mask=m > 0):
                            # sample spot variables
                            height = pyro.sample(
                                f"height_k{kdx}",
                                dist.HalfNormal(self.priors["height_std"]),
                            )
                            width = pyro.sample(
                                f"width_k{kdx}",
                                AffineBeta(
                                    1.5,
                                    2,
                                    self.priors["width_min"],
                                    self.priors["width_max"],
                                ),
                            )
                            x = pyro.sample(
                                f"x_k{kdx}",
                                AffineBeta(
                                    0,
                                    Vindex(size)[..., specific],
                                    -(self.data.P + 1) / 2,
                                    (self.data.P + 1) / 2,
                                ),
                            )
                            y = pyro.sample(
                                f"y_k{kdx}",
                                AffineBeta(
                                    0,
                                    Vindex(size)[..., specific],
                                    -(self.data.P + 1) / 2,
                                    (self.data.P + 1) / 2,
                                ),
                            )

                        # append
                        ms.append(m)
                        heights.append(height)
                        widths.append(width)
                        xs.append(x)
                        ys.append(y)

                    # observed data
                    pyro.sample(
                        "data",
                        KSMOGN(
                            torch.stack(heights, -1),
                            torch.stack(widths, -1),
                            torch.stack(xs, -1),
                            torch.stack(ys, -1),
                            target_locs,
                            background,
                            gain,
                            self.data.offset.samples,
                            self.data.offset.logits.to(self.dtype),
                            self.data.P,
                            torch.stack(torch.broadcast_tensors(*ms), -1),
                            use_pykeops=self.use_pykeops,
                        ),
                        obs=obs,
                    )

    def guide(self):
        r"""
        **Variational Distribution**

        .. math::
            \begin{aligned}
                q(\phi \setminus \{z, \theta\}) =~&q(g) q(\sigma^{xy}) q(\pi) q(\lambda) \cdot \\
                &\prod_{\mathsf{AOI}} \left[ q(\mu^b) q(\sigma^b) \prod_{\mathsf{frame}}
                \left[ \vphantom{\prod_{F}} q(b) \prod_{\mathsf{spot}}
                q(m) q(h | m) q(w | m) q(x | m) q(y | m) \right] \right]
            \end{aligned}
        """
        # global parameters
        pyro.sample(
            "gain",
            dist.Gamma(
                pyro.param("gain_loc") * pyro.param("gain_beta"),
                pyro.param("gain_beta"),
            ),
        )
        pyro.sample("pi", dist.Dirichlet(pyro.param("pi_mean") * pyro.param("pi_size")))
        pyro.sample(
            "lamda",
            dist.Gamma(
                pyro.param("lamda_loc") * pyro.param("lamda_beta"),
                pyro.param("lamda_beta"),
            ),
        )
        pyro.sample(
            "proximity",
            AffineBeta(
                pyro.param("proximity_loc"),
                pyro.param("proximity_size"),
                0,
                (self.data.P + 1) / math.sqrt(12),
            ),
        )

        # spots
        spots = pyro.plate("spots", self.K)
        # aoi sites
        aois = pyro.plate(
            "aois",
            self.data.Nt,
            subsample=self.n,
            subsample_size=self.nbatch_size,
            dim=-2,
        )
        # time frames
        frames = pyro.plate(
            "frames",
            self.data.F,
            subsample=self.f,
            subsample_size=self.fbatch_size,
            dim=-1,
        )

        with aois as ndx:
            ndx = ndx[:, None]
            mask = Vindex(self.data.mask)[ndx].to(self.device)
            with handlers.mask(mask=mask):
                pyro.sample(
                    "background_mean",
                    dist.Delta(Vindex(pyro.param("background_mean_loc"))[ndx, 0]),
                )
                pyro.sample(
                    "background_std",
                    dist.Delta(Vindex(pyro.param("background_std_loc"))[ndx, 0]),
                )
                with frames as fdx:
                    # sample background intensity
                    pyro.sample(
                        "background",
                        dist.Gamma(
                            Vindex(pyro.param("b_loc"))[ndx, fdx]
                            * Vindex(pyro.param("b_beta"))[ndx, fdx],
                            Vindex(pyro.param("b_beta"))[ndx, fdx],
                        ),
                    )

                    for kdx in spots:
                        # sample spot presence m
                        m = pyro.sample(
                            f"m_k{kdx}",
                            dist.Bernoulli(
                                Vindex(pyro.param("m_probs"))[kdx, ndx, fdx]
                            ),
                            infer={"enumerate": "parallel"},
                        )
                        with handlers.mask(mask=m > 0):
                            # sample spot variables
                            pyro.sample(
                                f"height_k{kdx}",
                                dist.Gamma(
                                    Vindex(pyro.param("h_loc"))[kdx, ndx, fdx]
                                    * Vindex(pyro.param("h_beta"))[kdx, ndx, fdx],
                                    Vindex(pyro.param("h_beta"))[kdx, ndx, fdx],
                                ),
                            )
                            pyro.sample(
                                f"width_k{kdx}",
                                AffineBeta(
                                    Vindex(pyro.param("w_mean"))[kdx, ndx, fdx],
                                    Vindex(pyro.param("w_size"))[kdx, ndx, fdx],
                                    0.75,
                                    2.25,
                                ),
                            )
                            pyro.sample(
                                f"x_k{kdx}",
                                AffineBeta(
                                    Vindex(pyro.param("x_mean"))[kdx, ndx, fdx],
                                    Vindex(pyro.param("size"))[kdx, ndx, fdx],
                                    -(self.data.P + 1) / 2,
                                    (self.data.P + 1) / 2,
                                ),
                            )
                            pyro.sample(
                                f"y_k{kdx}",
                                AffineBeta(
                                    Vindex(pyro.param("y_mean"))[kdx, ndx, fdx],
                                    Vindex(pyro.param("size"))[kdx, ndx, fdx],
                                    -(self.data.P + 1) / 2,
                                    (self.data.P + 1) / 2,
                                ),
                            )

    def init_parameters(self):
        """
        Initialize variational parameters.
        """
        device = self.device
        data = self.data

        pyro.param(
            "pi_mean",
            lambda: torch.ones(self.S + 1, device=device),
            constraint=constraints.simplex,
        )
        pyro.param(
            "pi_size",
            lambda: torch.tensor(2, device=device),
            constraint=constraints.positive,
        )
        pyro.param(
            "m_probs",
            lambda: torch.full((self.K, data.Nt, data.F), 0.5, device=device),
            constraint=constraints.unit_interval,
        )

        self._init_parameters()

    def _init_parameters(self):
        """
        Parameters shared between different models.
        """
        device = self.device
        data = self.data

        pyro.param(
            "proximity_loc",
            lambda: torch.tensor(0.5, device=device),
            constraint=constraints.interval(
                0,
                (self.data.P + 1) / math.sqrt(12) - torch.finfo(self.dtype).eps,
            ),
        )
        pyro.param(
            "proximity_size",
            lambda: torch.tensor(100, device=device),
            constraint=constraints.greater_than(2.0),
        )
        pyro.param(
            "lamda_loc",
            lambda: torch.tensor(0.5, device=device),
            constraint=constraints.positive,
        )
        pyro.param(
            "lamda_beta",
            lambda: torch.tensor(100, device=device),
            constraint=constraints.positive,
        )
        pyro.param(
            "gain_loc",
            lambda: torch.tensor(5, device=device),
            constraint=constraints.positive,
        )
        pyro.param(
            "gain_beta",
            lambda: torch.tensor(100, device=device),
            constraint=constraints.positive,
        )

        pyro.param(
            "background_mean_loc",
            lambda: torch.full(
                (data.Nt, 1),
                data.median[self.cdx] - self.data.offset.mean,
                device=device,
            ),
            constraint=constraints.positive,
        )
        pyro.param(
            "background_std_loc",
            lambda: torch.ones(data.Nt, 1, device=device),
            constraint=constraints.positive,
        )

        pyro.param(
            "b_loc",
            lambda: torch.full(
                (data.Nt, data.F),
                data.median[self.cdx] - self.data.offset.mean,
                device=device,
            ),
            constraint=constraints.positive,
        )
        pyro.param(
            "b_beta",
            lambda: torch.ones(data.Nt, data.F, device=device),
            constraint=constraints.positive,
        )
        pyro.param(
            "h_loc",
            lambda: torch.full((self.K, data.Nt, data.F), 2000, device=device),
            constraint=constraints.positive,
        )
        pyro.param(
            "h_beta",
            lambda: torch.full((self.K, data.Nt, data.F), 0.001, device=device),
            constraint=constraints.positive,
        )
        pyro.param(
            "w_mean",
            lambda: torch.full((self.K, data.Nt, data.F), 1.5, device=device),
            constraint=constraints.interval(
                0.75 + torch.finfo(self.dtype).eps,
                2.25 - torch.finfo(self.dtype).eps,
            ),
        )
        pyro.param(
            "w_size",
            lambda: torch.full((self.K, data.Nt, data.F), 100, device=device),
            constraint=constraints.greater_than(2.0),
        )
        pyro.param(
            "x_mean",
            lambda: torch.zeros(self.K, data.Nt, data.F, device=device),
            constraint=constraints.interval(
                -(data.P + 1) / 2 + torch.finfo(self.dtype).eps,
                (data.P + 1) / 2 - torch.finfo(self.dtype).eps,
            ),
        )
        pyro.param(
            "y_mean",
            lambda: torch.zeros(self.K, data.Nt, data.F, device=device),
            constraint=constraints.interval(
                -(data.P + 1) / 2 + torch.finfo(self.dtype).eps,
                (data.P + 1) / 2 - torch.finfo(self.dtype).eps,
            ),
        )
        pyro.param(
            "size",
            lambda: torch.full((self.K, data.Nt, data.F), 200, device=device),
            constraint=constraints.greater_than(2.0),
        )

    def TraceELBO(self, jit=False):
        """
        A trace implementation of ELBO-based SVI that supports - exhaustive enumeration over
        discrete sample sites, and - local parallel sampling over any sample site in the guide.
        """
        return (infer.JitTraceEnum_ELBO if jit else infer.TraceEnum_ELBO)(
            max_plate_nesting=2, ignore_jit_warnings=True
        )

    @lazy_property
    def compute_probs(self) -> torch.Tensor:
        z_probs = torch.zeros(self.data.Nt, self.data.F)
        theta_probs = torch.zeros(self.K, self.data.Nt, self.data.F)
        nbatch_size = self.nbatch_size
        fbatch_size = self.fbatch_size
        N = sum(self.data.is_ontarget)
        params = ["m", "x", "y"]
        params = list(map(lambda x: [f"{x}_k{i}" for i in range(self.K)], params))
        params = list(itertools.chain(*params))
        params += ["z", "theta"]
        theta_dims = tuple(i for i in range(0, self.Q * 2, 2))
        z_dims = tuple(i for i in range(1, self.Q * 2, 2))
        m_dims = tuple(i for i in range(self.Q * 2, self.Q * (self.K + 2)))
        for ndx in torch.split(torch.arange(N), nbatch_size):
            for fdx in torch.split(torch.arange(self.data.F), fbatch_size):
                self.n = ndx
                self.f = fdx
                self.nbatch_size = len(ndx)
                self.fbatch_size = len(fdx)
                with torch.no_grad(), pyro.plate(
                    "particles", size=50, dim=-3
                ), handlers.enum(first_available_dim=-4):
                    guide_tr = handlers.trace(self.guide).get_trace()
                    model_tr = handlers.trace(
                        handlers.replay(
                            handlers.block(self.model, hide=["data"]), trace=guide_tr
                        )
                    ).get_trace()
                model_tr.compute_log_prob()
                guide_tr.compute_log_prob()
                # 0 - theta
                # 1 - z
                # 2 - m_1
                # 3 - m_0
                # p(z, theta, phi)
                logp = 0
                for name in params:
                    logp = logp + model_tr.nodes[name]["unscaled_log_prob"]
                # p(z, theta | phi) = p(z, theta, phi) - p(z, theta, phi).sum(z, theta)
                logp = logp - logp.logsumexp(z_dims + theta_dims)
                m_log_probs = [
                    guide_tr.nodes[f"m_k{k}"]["unscaled_log_prob"]
                    for k in range(self.K)
                ]
                expectation = reduce(lambda x, y: x + y, m_log_probs) + logp
                # average over m
                result = expectation.logsumexp(m_dims)
                # marginalize theta
                z_logits = result.logsumexp(theta_dims)
                z_probs[ndx[:, None], fdx] = z_logits[1].exp().mean(-3)
                # marginalize z
                theta_logits = result.logsumexp(1)
                theta_probs[:, ndx[:, None], fdx] = theta_logits[1:].exp().mean(-3)
        self.n = None
        self.f = None
        self.nbatch_size = nbatch_size
        self.fbatch_size = fbatch_size
        return z_probs, theta_probs

    @property
    def z_probs(self) -> torch.Tensor:
        r"""
        Probability of there being a target-specific spot :math:`p(z=1)`
        """
        return self.compute_probs[0]

    @property
    def theta_probs(self) -> torch.Tensor:
        r"""
        Posterior target-specific spot probability :math:`q(\theta = k)`.
        """
        return self.compute_probs[1]

    @property
    def m_probs(self) -> torch.Tensor:
        r"""
        Posterior spot presence probability :math:`q(m=1)`.
        """
        return pyro.param("m_probs").data

    @property
    def pspecific(self) -> torch.Tensor:
        r"""
        Probability of there being a target-specific spot :math:`p(\mathsf{specific})`
        """
        return self.z_probs

    @property
    def z_map(self) -> torch.Tensor:
        return self.z_probs > 0.5



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

"""
hmm
^^^
"""

import math
from typing import Union

import funsor
import torch
import torch.distributions.constraints as constraints
from pyro.distributions.hmm import _logmatmulexp
from pyro.ops.indexing import Vindex
from pyroapi import distributions as dist
from pyroapi import handlers, infer, pyro
from torch.distributions.utils import lazy_property
from torch.nn.functional import one_hot

from tapqir.distributions import KSMOGN, AffineBeta
from tapqir.distributions.util import expand_offtarget, probs_m, probs_theta
from tapqir.models.cosmos import cosmos


class HMM(cosmos):
    r"""
    **Single-Color Hidden Markov Colocalization Model**

    EXPERIMENTAL This model relies on Funsor backend.

    :param S: Number of distinct molecular states for the binder molecules.
    :param K: Maximum number of spots that can be present in a single image.
    :param channels: Number of color channels.
    :param device: Computation device (cpu or gpu).
    :param dtype: Floating point precision.
    :param use_pykeops: Use pykeops as backend to marginalize out offset.
    :param vectorized: Vectorize time-dimension.
    """

    name = "hmm"

    def __init__(
        self,
        K: int = 2,
        channels: Union[tuple, list] = (0,),
        device: str = "cpu",
        dtype: str = "double",
        use_pykeops: bool = True,
        vectorized: bool = True,
        priors: dict = {
            "background_mean_std": 1000,
            "background_std_std": 100,
            "lamda_rate": 1,
            "height_std": 10000,
            "width_min": 0.75,
            "width_max": 2.25,
            "proximity_rate": 1,
            "gain_std": 50,
        },
    ):
        self.vectorized = vectorized
        super().__init__(K, channels, device, dtype, use_pykeops, priors)
        self._global_params = ["gain", "proximity", "lamda", "trans"]

    def model(self):
        """
        **Generative Model**
        """
        # global parameters
        gain = pyro.sample("gain", dist.HalfNormal(self.priors["gain_std"]))
        init = pyro.sample(
            "init", dist.Dirichlet(torch.ones(self.S + 1) / (self.S + 1))
        )
        init = expand_offtarget(init)
        trans = pyro.sample(
            "trans",
            dist.Dirichlet(torch.ones(self.S + 1, self.S + 1) / (self.S + 1)).to_event(
                1
            ),
        )
        trans = expand_offtarget(trans)
        lamda = pyro.sample("lamda", dist.Exponential(self.priors["lamda_rate"]))
        proximity = pyro.sample(
            "proximity", dist.Exponential(self.priors["proximity_rate"])
        )
        size = torch.stack(
            (
                torch.full_like(proximity, 2.0),
                (((self.data.P + 1) / (2 * proximity)) ** 2 - 1),
            ),
            dim=-1,
        )

        # spots
        spots = pyro.plate("spots", self.K)
        # aoi sites
        aois = pyro.plate(
            "aois",
            self.data.Nt,
            subsample=self.n,
            subsample_size=self.nbatch_size,
            dim=-2,
        )
        # time frames
        frames = (
            pyro.vectorized_markov(name="frames", size=self.data.F, dim=-1)
            if self.vectorized
            else pyro.markov(range(self.data.F))
        )

        with aois as ndx:
            ndx = ndx[:, None]
            # background mean and std
            background_mean = pyro.sample(
                "background_mean",
                dist.HalfNormal(self.priors["background_mean_std"]),
            )
            background_std = pyro.sample(
                "background_std", dist.HalfNormal(self.priors["background_std_std"])
            )
            z_prev = None
            for fdx in frames:
                if self.vectorized:
                    fsx, fdx = fdx
                else:
                    fsx = fdx
                # fetch data
                obs, target_locs, is_ontarget = self.data.fetch(ndx, fdx, self.cdx)
                # sample background intensity
                background = pyro.sample(
                    f"background_{fsx}",
                    dist.Gamma(
                        (background_mean / background_std) ** 2,
                        background_mean / background_std**2,
                    ),
                )

                # sample hidden model state (1+S,)
                z_probs = (
                    Vindex(init)[..., :, is_ontarget.long()]
                    if isinstance(fdx, int) and fdx < 1
                    else Vindex(trans)[..., z_prev, :, is_ontarget.long()]
                )
                z_curr = pyro.sample(f"z_{fsx}", dist.Categorical(z_probs))

                theta = pyro.sample(
                    f"theta_{fsx}",
                    dist.Categorical(
                        Vindex(probs_theta(self.K, self.device))[
                            torch.clamp(z_curr, min=0, max=1)
                        ]
                    ),
                    infer={"enumerate": "parallel"},
                )
                onehot_theta = one_hot(theta, num_classes=1 + self.K)

                ms, heights, widths, xs, ys = [], [], [], [], []
                for kdx in spots:
                    specific = onehot_theta[..., 1 + kdx]
                    # spot presence
                    m_probs = Vindex(probs_m(lamda, self.K))[..., theta, kdx]
                    m = pyro.sample(
                        f"m_{kdx}_{fsx}",
                        dist.Categorical(torch.stack((1 - m_probs, m_probs), -1)),
                    )
                    with handlers.mask(mask=m > 0):
                        # sample spot variables
                        height = pyro.sample(
                            f"height_{kdx}_{fsx}",
                            dist.HalfNormal(self.priors["height_std"]),
                        )
                        width = pyro.sample(
                            f"width_{kdx}_{fsx}",
                            AffineBeta(
                                1.5,
                                2,
                                self.priors["width_min"],
                                self.priors["width_max"],
                            ),
                        )
                        x = pyro.sample(
                            f"x_{kdx}_{fsx}",
                            AffineBeta(
                                0,
                                Vindex(size)[..., specific],
                                -(self.data.P + 1) / 2,
                                (self.data.P + 1) / 2,
                            ),
                        )
                        y = pyro.sample(
                            f"y_{kdx}_{fsx}",
                            AffineBeta(
                                0,
                                Vindex(size)[..., specific],
                                -(self.data.P + 1) / 2,
                                (self.data.P + 1) / 2,
                            ),
                        )

                    # append
                    ms.append(m)
                    heights.append(height)
                    widths.append(width)
                    xs.append(x)
                    ys.append(y)

                # observed data
                pyro.sample(
                    f"data_{fsx}",
                    KSMOGN(
                        torch.stack(heights, -1),
                        torch.stack(widths, -1),
                        torch.stack(xs, -1),
                        torch.stack(ys, -1),
                        target_locs,
                        background,
                        gain,
                        self.data.offset.samples,
                        self.data.offset.logits.to(self.dtype),
                        self.data.P,
                        torch.stack(torch.broadcast_tensors(*ms), -1),
                        use_pykeops=self.use_pykeops,
                    ),
                    obs=obs,
                )
                z_prev = z_curr

    def guide(self):
        """
        **Variational Distribution**
        """
        # global parameters
        pyro.sample(
            "gain",
            dist.Gamma(
                pyro.param("gain_loc") * pyro.param("gain_beta"),
                pyro.param("gain_beta"),
            ),
        )
        pyro.sample(
            "init", dist.Dirichlet(pyro.param("init_mean") * pyro.param("init_size"))
        )
        pyro.sample(
            "trans",
            dist.Dirichlet(
                pyro.param("trans_mean") * pyro.param("trans_size")
            ).to_event(1),
        )
        pyro.sample(
            "lamda",
            dist.Gamma(
                pyro.param("lamda_loc") * pyro.param("lamda_beta"),
                pyro.param("lamda_beta"),
            ),
        )
        pyro.sample(
            "proximity",
            AffineBeta(
                pyro.param("proximity_loc"),
                pyro.param("proximity_size"),
                0,
                (self.data.P + 1) / math.sqrt(12),
            ),
        )

        # spots
        spots = pyro.plate("spots", self.K)
        # aoi sites
        aois = pyro.plate(
            "aois",
            self.data.Nt,
            subsample=self.n,
            subsample_size=self.nbatch_size,
            dim=-2,
        )
        # time frames
        frames = (
            pyro.vectorized_markov(name="frames", size=self.data.F, dim=-1)
            if self.vectorized
            else pyro.markov(range(self.data.F))
        )

        with aois as ndx:
            ndx = ndx[:, None]
            pyro.sample(
                "background_mean",
                dist.Delta(Vindex(pyro.param("background_mean_loc"))[ndx, 0]),
            )
            pyro.sample(
                "background_std",
                dist.Delta(Vindex(pyro.param("background_std_loc"))[ndx, 0]),
            )
            z_prev = None
            for fdx in frames:
                if self.vectorized:
                    fsx, fdx = fdx
                else:
                    fsx = fdx
                # sample background intensity
                pyro.sample(
                    f"background_{fsx}",
                    dist.Gamma(
                        Vindex(pyro.param("b_loc"))[ndx, fdx]
                        * Vindex(pyro.param("b_beta"))[ndx, fdx],
                        Vindex(pyro.param("b_beta"))[ndx, fdx],
                    ),
                )

                # sample hidden model state
                z_probs = (
                    Vindex(pyro.param("z_trans"))[ndx, fdx, 0]
                    if isinstance(fdx, int) and fdx < 1
                    else Vindex(pyro.param("z_trans"))[ndx, fdx, z_prev]
                )
                z_curr = pyro.sample(
                    f"z_{fsx}",
                    dist.Categorical(z_probs),
                    infer={"enumerate": "parallel"},
                )

                for kdx in spots:
                    # spot presence
                    m_probs = Vindex(pyro.param("m_probs"))[z_curr, kdx, ndx, fdx]
                    m = pyro.sample(
                        f"m_{kdx}_{fsx}",
                        dist.Categorical(torch.stack((1 - m_probs, m_probs), -1)),
                        infer={"enumerate": "parallel"},
                    )
                    with handlers.mask(mask=m > 0):
                        # sample spot variables
                        pyro.sample(
                            f"height_{kdx}_{fsx}",
                            dist.Gamma(
                                Vindex(pyro.param("h_loc"))[kdx, ndx, fdx]
                                * Vindex(pyro.param("h_beta"))[kdx, ndx, fdx],
                                Vindex(pyro.param("h_beta"))[kdx, ndx, fdx],
                            ),
                        )
                        pyro.sample(
                            f"width_{kdx}_{fsx}",
                            AffineBeta(
                                Vindex(pyro.param("w_mean"))[kdx, ndx, fdx],
                                Vindex(pyro.param("w_size"))[kdx, ndx, fdx],
                                0.75,
                                2.25,
                            ),
                        )
                        pyro.sample(
                            f"x_{kdx}_{fsx}",
                            AffineBeta(
                                Vindex(pyro.param("x_mean"))[kdx, ndx, fdx],
                                Vindex(pyro.param("size"))[kdx, ndx, fdx],
                                -(self.data.P + 1) / 2,
                                (self.data.P + 1) / 2,
                            ),
                        )
                        pyro.sample(
                            f"y_{kdx}_{fsx}",
                            AffineBeta(
                                Vindex(pyro.param("y_mean"))[kdx, ndx, fdx],
                                Vindex(pyro.param("size"))[kdx, ndx, fdx],
                                -(self.data.P + 1) / 2,
                                (self.data.P + 1) / 2,
                            ),
                        )

                z_prev = z_curr

    def init_parameters(self):
        """
        Initialize variational parameters.
        """
        device = self.device
        data = self.data

        pyro.param(
            "init_mean",
            lambda: torch.ones(self.S + 1, device=device),
            constraint=constraints.simplex,
        )
        pyro.param(
            "init_size",
            lambda: torch.tensor(2, device=device),
            constraint=constraints.positive,
        )
        pyro.param(
            "trans_mean",
            lambda: torch.ones(self.S + 1, self.S + 1, device=device),
            constraint=constraints.simplex,
        )
        pyro.param(
            "trans_size",
            lambda: torch.full((self.S + 1, 1), 2, device=device),
            constraint=constraints.positive,
        )

        # classification
        pyro.param(
            "z_trans",
            lambda: torch.ones(
                data.Nt,
                data.F,
                1 + self.S,
                1 + self.S,
                device=device,
            ),
            constraint=constraints.simplex,
        )
        pyro.param(
            "m_probs",
            lambda: torch.full(
                (1 + self.S, self.K, self.data.Nt, self.data.F),
                0.5,
                device=device,
            ),
            constraint=constraints.unit_interval,
        )

        self._init_parameters()

    def TraceELBO(self, jit=False):
        """
        A trace implementation of ELBO-based SVI that supports - exhaustive enumeration over
        discrete sample sites, and - local parallel sampling over any sample site in the guide.
        """
        if self.vectorized:
            return (
                infer.JitTraceMarkovEnum_ELBO if jit else infer.TraceMarkovEnum_ELBO
            )(max_plate_nesting=2, ignore_jit_warnings=True)
        return (infer.JitTraceEnum_ELBO if jit else infer.TraceEnum_ELBO)(
            max_plate_nesting=2, ignore_jit_warnings=True
        )

    @staticmethod
    def _sequential_logmatmulexp(logits: torch.Tensor) -> torch.Tensor:
        """
        For a tensor ``x`` whose time dimension is -3, computes::
            x[..., 0, :, :] @ x[..., 1, :, :] @ ... @ x[..., T-1, :, :]
        but does so numerically stably in log space.
        """
        batch_shape = logits.shape[:-3]
        state_dim = logits.size(-1)
        sum_terms = []
        # up sweep
        while logits.size(-3) > 1:
            time = logits.size(-3)
            even_time = time // 2 * 2
            even_part = logits[..., :even_time, :, :]
            x_y = even_part.reshape(
                batch_shape + (even_time // 2, 2, state_dim, state_dim)
            )
            x, y = x_y.unbind(-3)
            contracted = _logmatmulexp(x, y)
            if time > even_time:
                contracted = torch.cat((contracted, logits[..., -1:, :, :]), dim=-3)
            sum_terms.append(logits)
            logits = contracted
        else:
            sum_terms.append(logits)
        # handle root case
        sum_term = sum_terms.pop()
        left_term = HMM._contraction_identity(sum_term)
        # down sweep
        while sum_terms:
            sum_term = sum_terms.pop()
            new_left_term = HMM._contraction_identity(sum_term)
            time = sum_term.size(-3)
            even_time = time // 2 * 2
            if time > even_time:
                new_left_term[..., time - 1 : time, :, :] = left_term[
                    ..., even_time // 2 : even_time // 2 + 1, :, :
                ]
                left_term = left_term[..., : even_time // 2, :, :]

            left_sum = sum_term[..., :even_time:2, :, :]
            left_sum_and_term = _logmatmulexp(left_term, left_sum)
            new_left_term[..., :even_time:2, :, :] = left_term
            new_left_term[..., 1:even_time:2, :, :] = left_sum_and_term
            left_term = new_left_term
        else:
            alphas = _logmatmulexp(left_term, sum_term)
        return alphas

    @staticmethod
    def _contraction_identity(logits: torch.Tensor) -> torch.Tensor:
        batch_shape = logits.shape[:-2]
        state_dim = logits.size(-1)
        result = torch.eye(state_dim).log()
        result = result.reshape((1,) * len(batch_shape) + (state_dim, state_dim))
        result = result.repeat(batch_shape + (1, 1))
        return result

    @lazy_property
    def compute_probs(self) -> torch.Tensor:
        theta_probs = torch.zeros(self.K, self.data.Nt, self.data.F)
        nbatch_size = self.nbatch_size
        N = sum(self.data.is_ontarget)
        for ndx in torch.split(torch.arange(N), nbatch_size):
            self.n = ndx
            self.nbatch_size = len(ndx)
            with torch.no_grad(), pyro.plate(
                "particles", size=5, dim=-3
            ), handlers.enum(first_available_dim=-4):
                guide_tr = handlers.trace(self.guide).get_trace()
                model_tr = handlers.trace(
                    handlers.replay(self.model, trace=guide_tr)
                ).get_trace()
            model_tr.compute_log_prob()
            guide_tr.compute_log_prob()

            logp = {}
            result = {}
            for fsx in ("0", f"slice(1, {self.data.F}, None)"):
                logp[fsx] = 0
                # collect log_prob terms p(z, theta, phi)
                for name in ["z", "theta", "m_0", "m_1", "x_0", "x_1", "y_0", "y_1"]:
                    logp[fsx] += model_tr.nodes[f"{name}_{fsx}"]["funsor"]["log_prob"]
                if fsx == "0":
                    # substitute MAP values of z into p(z=z_map, theta, phi)
                    z_map = funsor.Tensor(self.z_map[ndx, 0].long(), dtype=2)["aois"]
                    logp[fsx] = logp[fsx](**{f"z_{fsx}": z_map})
                    # compute log_measure q for given z_map
                    log_measure = guide_tr.nodes[f"m_0_{fsx}"]["funsor"]["log_measure"]
                    +guide_tr.nodes[f"m_1_{fsx}"]["funsor"]["log_measure"]
                    log_measure = log_measure(**{f"z_{fsx}": z_map})
                else:
                    # substitute MAP values of z into p(z=z_map, theta, phi)
                    z_map = funsor.Tensor(self.z_map[ndx, 1:].long(), dtype=2)[
                        "aois", "frames"
                    ]
                    z_map_prev = funsor.Tensor(self.z_map[ndx, :-1].long(), dtype=2)[
                        "aois", "frames"
                    ]
                    fsx_prev = f"slice(0, {self.data.F-1}, None)"
                    logp[fsx] = logp[fsx](
                        **{f"z_{fsx}": z_map, f"z_{fsx_prev}": z_map_prev}
                    )
                    # compute log_measure q for given z_map
                    log_measure = guide_tr.nodes[f"m_0_{fsx}"]["funsor"]["log_measure"]
                    +guide_tr.nodes[f"m_1_{fsx}"]["funsor"]["log_measure"]
                    log_measure = log_measure(
                        **{f"z_{fsx}": z_map, f"z_{fsx_prev}": z_map_prev}
                    )
                # compute p(z_map, theta | phi) = p(z_map, theta, phi) - p(z_map, phi)
                logp[fsx] = logp[fsx] - logp[fsx].reduce(
                    funsor.ops.logaddexp, f"theta_{fsx}"
                )
                # average over m in p * q
                result[fsx] = (logp[fsx] + log_measure).reduce(
                    funsor.ops.logaddexp, frozenset({f"m_0_{fsx}", f"m_1_{fsx}"})
                )
                # average over particles
                result[fsx] = result[fsx].exp().reduce(funsor.ops.mean, "particles")
            theta_probs[:, ndx, 0] = result["0"].data[:, 1:].permute(1, 0)
            theta_probs[:, ndx, 1:] = (
                result[f"slice(1, {self.data.F}, None)"].data[..., 1:].permute(2, 0, 1)
            )
        self.n = None
        self.nbatch_size = nbatch_size
        return theta_probs

    @property
    def z_probs(self) -> torch.Tensor:
        r"""
        Probability of there being a target-specific spot :math:`p(z=1)`
        """
        result = self._sequential_logmatmulexp(pyro.param("z_trans").data.log())
        return result[..., 0, 1].exp()

    @property
    def theta_probs(self) -> torch.Tensor:
        r"""
        Posterior target-specific spot probability :math:`q(\theta = k, z=z_\mathsf{MAP})`.
        """
        return self.compute_probs

    @property
    def pspecific(self) -> torch.Tensor:
        r"""
        Probability of there being a target-specific spot :math:`p(\mathsf{specific})`
        """
        return self.z_probs

    @property
    def m_probs(self) -> torch.Tensor:
        r"""
        Posterior spot presence probability :math:`q(m=1, z=z_\mathsf{MAP})`.
        """
        return Vindex(torch.permute(pyro.param("m_probs").data, (1, 2, 3, 0)))[
            ..., self.z_map.long()
        ]



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import logging
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd
import torch
from pyro.ops.indexing import Vindex
from pyro.ops.stats import hpdi, quantile
from sklearn.metrics import (
    confusion_matrix,
    matthews_corrcoef,
    precision_score,
    recall_score,
)

from tapqir.distributions.util import gaussian_spots
from tapqir.handlers import StatsMessenger

logger = logging.getLogger(__name__)


def snr_and_chi2(
    data: torch.Tensor,
    height: torch.Tensor,
    width: torch.Tensor,
    x: torch.Tensor,
    y: torch.Tensor,
    target_locs: torch.Tensor,
    background: torch.Tensor,
    gain: float,
    offset_mean: float,
    offset_var: float,
    P: int,
    theta_probs: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""
    Calculate the signal-to-noise ratio.

    Total signal:

    .. math::
        \mu_{knf} =  \sum_{ij} I_{nfij}
        \mathcal{N}(i, j \mid x_{knf}, y_{knf}, w_{knf})`

    Noise:

    .. math::
        \sigma^2_{knf} = \sigma^2_{\text{offset}}
        + \mu_{knf} \text{gain}`

    Signal-to-noise ratio:

    .. math::
        \text{SNR}_{knf} =
        \dfrac{\mu_{knf} - b_{nf} - \mu_{\text{offset}}}{\sigma_{knf}}
        \text{ for } \theta_{nf} = k`
    """
    gaussians = gaussian_spots(
        height,
        width,
        x,
        y,
        target_locs,
        P,
    )

    # snr
    weights = gaussians / height[..., None, None]
    signal = ((data - background[..., None, None] - offset_mean) * weights).sum(
        dim=(-2, -1)
    )
    noise = (offset_var + background * gain).sqrt()
    snr_result = signal / noise

    # chi2 test
    img_ideal = background[..., None, None] + gaussians.sum(-4)
    chi2_result = (data - img_ideal - offset_mean) ** 2 / img_ideal

    return snr_result, chi2_result.mean(dim=(-1, -2))


def save_stats(model, path, CI=0.95, save_matlab=False):
    # global parameters
    global_params = model._global_params
    summary = pd.DataFrame(
        index=global_params,
        columns=["Mean", f"{int(100*CI)}% LL", f"{int(100*CI)}% UL"],
    )

    logger.info("- credible intervals")
    nbatch_size = model.nbatch_size
    fbatch_size = model.fbatch_size
    model.nbatch_size = None
    model.fbatch_size = None
    with StatsMessenger(
        CI=CI, K=model.K, N=model.data.Nt, F=model.data.F, Q=model.Q
    ) as ci_stats:
        model.guide()
    model.nbatch_size = nbatch_size
    model.fbatch_size = fbatch_size

    for param in global_params:
        if ci_stats[param]["Mean"].ndim == 0:
            summary.loc[param, "Mean"] = ci_stats[param]["Mean"].item()
            summary.loc[param, "95% LL"] = ci_stats[param]["LL"].item()
            summary.loc[param, "95% UL"] = ci_stats[param]["UL"].item()
        else:
            summary.loc[param, "Mean"] = ci_stats[param]["Mean"].tolist()
            summary.loc[param, "95% LL"] = ci_stats[param]["LL"].tolist()
            summary.loc[param, "95% UL"] = ci_stats[param]["UL"].tolist()
    logger.info("- spot probabilities")
    ci_stats["m_probs"] = model.m_probs.data.cpu()
    ci_stats["theta_probs"] = model.theta_probs.data.cpu()
    ci_stats["z_probs"] = model.z_probs.data.cpu()
    ci_stats["z_map"] = model.z_map.data.cpu()
    ci_stats["p_specific"] = ci_stats["theta_probs"].sum(0)

    # calculate vmin/vmax
    theta_mask = ci_stats["theta_probs"] > 0.5
    if theta_mask.sum():
        hmax = np.percentile(ci_stats["height"]["Mean"][theta_mask], 99)
    else:
        hmax = 1
    ci_stats["height"]["vmin"] = -0.03 * hmax
    ci_stats["height"]["vmax"] = 1.3 * hmax
    ci_stats["width"]["vmin"] = 0.5
    ci_stats["width"]["vmax"] = 2.5
    ci_stats["x"]["vmin"] = -9
    ci_stats["x"]["vmax"] = 9
    ci_stats["y"]["vmin"] = -9
    ci_stats["y"]["vmax"] = 9
    bmax = np.percentile(ci_stats["background"]["Mean"].flatten(), 99)
    ci_stats["background"]["vmin"] = -0.03 * bmax
    ci_stats["background"]["vmax"] = 1.3 * bmax

    # timestamps
    if model.data.time1 is not None:
        ci_stats["time1"] = model.data.time1
    if model.data.ttb is not None:
        ci_stats["ttb"] = model.data.ttb

    # intensity of target-specific spots
    theta_mask = torch.argmax(ci_stats["theta_probs"], dim=0)
    h_specific = Vindex(ci_stats["height"]["Mean"])[
        theta_mask, torch.arange(model.data.Nt)[:, None], torch.arange(model.data.F)
    ]
    ci_stats["h_specific"] = h_specific * (ci_stats["z_map"] > 0).long()

    model.params = ci_stats

    logger.info("- SNR and Chi2-test")
    # snr and chi2 test
    snr = torch.zeros(model.K, model.data.Nt, model.data.F, device=torch.device("cpu"))
    chi2 = torch.zeros(model.data.Nt, model.data.F, device=torch.device("cpu"))
    for n in range(model.data.Nt):
        snr[:, n], chi2[n] = snr_and_chi2(
            model.data.images[n, :, model.cdx],
            ci_stats["height"]["Mean"][:, n],
            ci_stats["width"]["Mean"][:, n],
            ci_stats["x"]["Mean"][:, n],
            ci_stats["y"]["Mean"][:, n],
            model.data.xy[n, :, model.cdx],
            ci_stats["background"]["Mean"][n],
            ci_stats["gain"]["Mean"],
            model.data.offset.mean,
            model.data.offset.var,
            model.data.P,
            ci_stats["theta_probs"][:, n],
        )
    snr_masked = snr[ci_stats["theta_probs"] > 0.5]
    summary.loc["SNR", "Mean"] = snr_masked.mean().item()
    ci_stats["chi2"] = {}
    ci_stats["chi2"]["values"] = chi2
    cmax = quantile(ci_stats["chi2"]["values"].flatten(), 0.99)
    ci_stats["chi2"]["vmin"] = -0.03 * cmax
    ci_stats["chi2"]["vmax"] = 1.3 * cmax

    # classification statistics
    if model.data.labels is not None:
        pred_labels = model.z_map[model.data.is_ontarget].cpu().numpy().ravel()
        true_labels = model.data.labels["z"][: model.data.N, :, model.cdx].ravel()

        with np.errstate(divide="ignore", invalid="ignore"):
            summary.loc["MCC", "Mean"] = matthews_corrcoef(true_labels, pred_labels)
        summary.loc["Recall", "Mean"] = recall_score(
            true_labels, pred_labels, zero_division=0
        )
        summary.loc["Precision", "Mean"] = precision_score(
            true_labels, pred_labels, zero_division=0
        )

        (
            summary.loc["TN", "Mean"],
            summary.loc["FP", "Mean"],
            summary.loc["FN", "Mean"],
            summary.loc["TP", "Mean"],
        ) = confusion_matrix(true_labels, pred_labels, labels=(0, 1)).ravel()

        mask = torch.from_numpy(model.data.labels["z"][: model.data.N, :, model.cdx])
        samples = torch.masked_select(model.z_probs[model.data.is_ontarget].cpu(), mask)
        if len(samples):
            z_ll, z_ul = hpdi(samples, CI)
            summary.loc["p(specific)", "Mean"] = quantile(samples, 0.5).item()
            summary.loc["p(specific)", "95% LL"] = z_ll.item()
            summary.loc["p(specific)", "95% UL"] = z_ul.item()
        else:
            summary.loc["p(specific)", "Mean"] = 0.0
            summary.loc["p(specific)", "95% LL"] = 0.0
            summary.loc["p(specific)", "95% UL"] = 0.0

    model.summary = summary

    if path is not None:
        path = Path(path)
        param_path = path / f"{model.full_name}-params.tpqr"
        torch.save(ci_stats, param_path)
        logger.info(f"Parameters were saved in {param_path}")
        if save_matlab:
            from scipy.io import savemat

            for param, field in ci_stats.items():
                if param in (
                    "m_probs",
                    "theta_probs",
                    "z_probs",
                    "z_map",
                    "p_specific",
                    "h_specific",
                    "time1",
                    "ttb",
                ):
                    ci_stats[param] = np.asarray(field)
                    continue
                for stat, value in field.items():
                    ci_stats[param][stat] = np.asarray(value)
            mat_path = path / f"{model.full_name}-params.mat"
            savemat(mat_path, ci_stats)
            logger.info(f"Matlab parameters were saved in {mat_path}")
        csv_path = path / f"{model.full_name}-summary.csv"
        summary.to_csv(csv_path)
        logger.info(f"Summary statistics were saved in {csv_path}")



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import os

import setuptools

import versioneer

PROJECT_PATH = os.path.dirname(os.path.abspath(__file__))
VERSION = """
# This file is auto-generated with the version information during setup.py installation.
__version__ = '{}'
"""

with open("README.rst", "r") as fh:
    long_description = fh.read()

# examples/tutorials
EXTRAS_REQUIRE = [
    "notebook",
]
# tests
TEST_REQUIRE = [
    "black[jupyter]",
    "flake8",
    "isort",
    "pytest",
    "pytest-xvfb",
]
# docs
DOCS_REQUIRE = [
    "IPython",
    "nbsphinx>=0.8.5",
    "pydata_sphinx_theme",
    "sphinx",
    "sphinx-autodoc-typehints",
    "sphinx-click",
    "sphinx-copybutton",
    "sphinx-gallery",
    "sphinx-panels",
]
DESKTOP_REQUIRE = ["voila"]

setuptools.setup(
    name="tapqir",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Yerdos Ordabayev",
    author_email="ordabayev@brandeis.edu",
    description="Bayesian analysis of co-localization single-molecule microscopy image data",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://tapqir.readthedocs.io",
    project_urls={
        "Documentation": "https://tapqir.readthedocs.io",
        "Source": "https://github.com/gelles-brandeis/tapqir",
    },
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        "colorama",
        "funsor==0.4.2",
        "future",
        "ipyevents",
        "ipyfilechooser",
        "ipympl",
        "ipywidgets",
        "matplotlib",
        "pandas",
        "pykeops>=2.0",
        "pyro-ppl==1.8.0",
        "pyyaml>=6.0",
        "scikit-learn",
        "scipy",
        "tensorboard",
        "torch==1.10.0",
        "typer",
    ],
    extras_require={
        "desktop": DESKTOP_REQUIRE,
        "extras": EXTRAS_REQUIRE,
        "test": EXTRAS_REQUIRE + TEST_REQUIRE,
        "docs": DOCS_REQUIRE,
        "dev": DESKTOP_REQUIRE + EXTRAS_REQUIRE + TEST_REQUIRE + DOCS_REQUIRE,
    },
    keywords="image-classification probabilistic-programming cosmos pyro",
    license="Apache 2.0",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.7",
    entry_points={
        "console_scripts": [
            "tapqir=tapqir.main:app",
            "tapqir-gui=tapqir.gui:app",
        ],
    },
)



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import logging

import colorama


class ColorFormatter(logging.Formatter):
    """
    The code adapted from https://github.com/iterative/dvc/blob/main/dvc/logger.py

    Spit out colored text in supported terminals.
    colorama__ makes ANSI escape character sequences work under Windows.
    See the colorama documentation for details.
    __ https://pypi.python.org/pypi/colorama
    If record has an extra `tb_only` attribute, it will not show the
    exception cause, just the message and the traceback.
    """

    color_code = {
        "TRACE": colorama.Fore.GREEN,
        "DEBUG": colorama.Fore.BLUE,
        "WARNING": colorama.Fore.YELLOW,
        "ERROR": colorama.Fore.RED,
        "CRITICAL": colorama.Fore.RED,
    }

    def format(self, record):
        record.message = record.getMessage()
        msg = self.formatMessage(record)

        if record.levelname == "INFO":
            return msg

        if record.exc_info:
            if getattr(record, "tb_only", False):
                cause = ""
            else:
                cause = ": ".join(_iter_causes(record.exc_info[1]))

            msg = "{message}{separator}{cause}".format(
                message=msg or "",
                separator=" - " if msg and cause else "",
                cause=cause,
            )

        return "{asctime}{color}{levelname}{nc}: {msg}".format(
            asctime=self.formatTime(record, self.datefmt),
            color=self.color_code[record.levelname],
            nc=colorama.Fore.RESET,
            levelname=record.levelname,
            msg=msg,
        )

    def formatTime(self, record, datefmt=None):
        return ""


def _iter_causes(exc):
    while exc:
        yield str(exc)
        exc = exc.__cause__


def _stack_trace(exc_info):
    import traceback

    return (
        "\n"
        "{red}{line}{nc}\n"
        "{trace}"
        "{red}{line}{nc}".format(
            red=colorama.Fore.RED,
            line="-" * 60,
            trace="".join(traceback.format_exception(*exc_info)),
            nc=colorama.Fore.RESET,
        )
    )



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import logging
import os
from collections import OrderedDict, defaultdict
from functools import partial, singledispatch
from pathlib import Path

import ipywidgets as widgets
import matplotlib.pyplot as plt
import torch
from ipyevents import Event
from ipyfilechooser import FileChooser
from IPython.display import display
from tensorboard import notebook
from tqdm import tqdm_notebook
from traitlets.utils.bunch import Bunch

from tapqir.distributions.util import gaussian_spots
from tapqir.main import avail_models, fit, glimpse, log, main, show
from tapqir.utils.dataset import save

logger = logging.getLogger("tapqir")

# disable pyplot keyboard shortcuts
plt.rcParams["keymap.zoom"].remove("o")
plt.rcParams["keymap.home"].remove("h")
plt.rcParams["keymap.yscale"].remove("l")
plt.rcParams["keymap.xscale"].remove("k")

avail_models = list(avail_models.__members__.keys())


@singledispatch
def get_value(x):
    return x


@get_value.register
def _(x: Bunch):
    return x.new


@get_value.register
def _(x: widgets.Widget):
    return x.value


@singledispatch
def get_path(x):
    return x


@get_path.register
def _(x: str):
    return Path(x)


@get_path.register
def _(x: FileChooser):
    return Path(x.selected_path)


class inputBox(widgets.VBox):
    def __init__(self, **kwargs):
        self._children = OrderedDict()
        super().__init__(children=[], **kwargs)

    def build_children(self) -> None:
        self.children = [
            child["widget"] for child in self._children.values() if not child["hide"]
        ]

    def add_child(self, name, widget, hide=False, beginning=False) -> None:
        self._children[name] = {
            "widget": widget,
            "hide": hide,
        }
        if beginning:
            self._children.move_to_end(name, last=False)
        self.build_children()

    def remove_child(self, name) -> None:
        self._children.pop(name)
        self.build_children()

    def toggle_hide(self, a=None, names=None) -> None:
        for name in names:
            self._children[name]["hide"] = not self._children[name]["hide"]
        self.build_children()

    def __getitem__(self, name):
        return self._children[name]["widget"]

    @property
    def kwargs(self):
        result = {}
        for name, v in self._children.items():
            if isinstance(
                v["widget"],
                (
                    widgets.Button,
                    widgets.Output,
                    widgets.Accordion,
                    widgets.VBox,
                    widgets.HBox,
                ),
            ):
                continue
            if isinstance(v["widget"], widgets.Tab):
                keys = v["widget"].children[0]._children.keys()
                for k in keys:
                    result[k] = [get_value(c[k]) for c in v["widget"].children]
                continue
            result[name] = get_value(v["widget"])
            if name == "channels":  # FIXME
                result[name] = [int(result[name])]
        return result


def initUI(DEFAULTS):
    """
    Main layout.

    - Set working directory.
    - Commands tabs.
    - Output.
    """

    layout = inputBox(layout={"width": "850px", "border": "2px solid black"})

    # widgets
    cd = FileChooser(".", title="Select working directory")
    cd.show_only_dirs = True

    out = widgets.Output()

    # layout
    layout.add_child("cd", cd)
    layout.add_child("out", out)
    # callbacks
    cd.register_callback(
        partial(
            cdCmd,
            DEFAULTS=DEFAULTS,
            layout=layout,
        )
    )
    return layout


def cdCmd(path, DEFAULTS, layout):
    """
    Set working directory and load default parameters (main).
    """
    import sys
    from platform import uname

    IN_COLAB = "google.colab" in sys.modules
    IN_WSL2 = "microsoft-standard" in uname().release
    IN_LINUX = not IN_COLAB and not IN_WSL2

    out = layout["out"]
    path = get_path(path)
    with out:
        main(cd=path)
        logger.info("Loading configuration data ...")

    # Tabs
    tab = widgets.Tab()
    tab.children = [glimpseUI(out, DEFAULTS), fitUI(out, DEFAULTS)]
    if IN_LINUX:
        tensorboard = widgets.Output()
        tab.children = tab.children + (showUI(out, DEFAULTS), tensorboard)
    elif IN_WSL2:
        tensorboard = None
        tensorboard_info = widgets.Label(
            value=(
                'Run "tensorboard --logdir=<working directory>" in the terminal'
                ' and then open "localhost:6006" in the browser'
            )
        )
        tab.children = tab.children + (showUI(out, DEFAULTS), tensorboard_info)
    elif IN_COLAB:
        tensorboard = None
        tab.children = tab.children + (
            widgets.Label(value="Disabled in Colab"),
            widgets.Label(value="Disabled in Colab"),
        )
    tab.children = tab.children + (logUI(out),)
    tab.set_title(0, "Extract AOIs")
    tab.set_title(1, "Fit the data")
    tab.set_title(2, "View results")
    tab.set_title(3, "Tensorboard")
    tab.set_title(4, "View logs")

    if tensorboard is not None:
        with tensorboard:
            notebook.start(f"--logdir '{path}'")
            notebook.display(height=1000)

    # insert tabs into GUI
    wd = widgets.Label(value=f"Working directory: {path}")
    layout.remove_child("cd")
    layout.add_child("tab", tab, beginning=True)
    layout.add_child("wd", wd, beginning=True)

    with out:
        logger.info("Loading configuration data: Done")

    out.clear_output(wait=True)


def glimpseUI(out, DEFAULTS):
    """
    Glimpse command layout.
    """
    layout = inputBox()
    layout.add_child(
        "dataset",
        widgets.Text(
            value=DEFAULTS["dataset"],
            description="Dataset name",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "P",
        widgets.BoundedIntText(
            value=DEFAULTS["P"],
            min=5,
            max=50,
            description="AOI image size",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "offset_x",
        widgets.IntText(
            value=DEFAULTS["offset-x"],
            description="Offset region top-left corner (x-axis)",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "offset_y",
        widgets.IntText(
            value=DEFAULTS["offset-y"],
            description="Offset region top-left corner (y-axis)",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "offset_P",
        widgets.IntText(
            value=DEFAULTS["offset-P"],
            description="Offset region size",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "bin_size",
        widgets.BoundedIntText(
            value=DEFAULTS["bin-size"],
            min=1,
            max=21,
            step=2,
            description="Offset histogram bin size (odd number)",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "frame_range",
        widgets.Checkbox(
            value=DEFAULTS["frame-range"],
            description="Specify frame range?",
            indent=False,
        ),
    )
    layout.add_child(
        "frame_start",
        widgets.IntText(
            value=DEFAULTS["frame-start"],
            description="Starting frame",
            style={"description_width": "initial"},
        ),
        hide=not DEFAULTS["frame-range"],
    )
    layout.add_child(
        "frame_end",
        widgets.IntText(
            value=DEFAULTS["frame-end"],
            description="Ending frame",
            style={"description_width": "initial"},
        ),
        hide=not DEFAULTS["frame-range"],
    )
    layout.add_child(
        "use_offtarget",
        widgets.Checkbox(
            value=DEFAULTS["use-offtarget"],
            description="Use off-target AOI locations?",
            indent=False,
        ),
    )
    layout.add_child(
        "num_channels",
        widgets.BoundedIntText(
            value=DEFAULTS["num-channels"],
            min=1,
            max=4,
            step=1,
            description="Number of color channels",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child("channels", widgets.Tab())
    channelUI(
        layout["num_channels"].value,
        layout["channels"],
        layout["use_offtarget"],
        DEFAULTS,
    )
    layout.add_child("extract_aois", widgets.Button(description="Extract AOIs"))
    # Callbacks
    layout["num_channels"].observe(
        partial(
            channelUI,
            channelTabs=layout["channels"],
            useOfftarget=layout["use_offtarget"],
            DEFAULTS=DEFAULTS,
        ),
        names="value",
    )
    layout["use_offtarget"].observe(
        partial(toggleUseOffTarget, channelTabs=layout["channels"]),
        names="value",
    )
    layout["extract_aois"].on_click(partial(glimpseCmd, layout=layout, out=out))
    layout["frame_range"].observe(
        partial(layout.toggle_hide, names=("frame_start", "frame_end")),
        names="value",
    )
    return layout


def toggleUseOffTarget(checked, channelTabs):
    for c, tab in enumerate(channelTabs.children):
        tab.toggle_hide(names=("offtarget_aoiinfo",))


def glimpseCmd(b, layout, out):
    with out:
        glimpse(
            **layout.kwargs,
            no_input=True,
            progress_bar=tqdm_notebook,
            labels=False,
        )

    out.clear_output(wait=True)


def channelUI(C, channelTabs, useOfftarget, DEFAULTS):
    C = get_value(C)
    currentC = len(channelTabs.children)
    for c in range(max(currentC, C)):
        if c < C and c < currentC:
            continue
        elif c < C and c >= currentC:
            if len(DEFAULTS["channels"]) < c + 1:
                DEFAULTS["channels"].append(defaultdict(lambda: None))
            layout = inputBox()
            layout.add_child(
                "name",
                widgets.Text(
                    value=DEFAULTS["channels"][c]["name"],
                    description="Channel name",
                    style={"description_width": "initial"},
                ),
            )
            layout.add_child(
                "glimpse_folder", FileChooser(".", title="Header/glimpse folder")
            )
            if DEFAULTS["channels"][c]["glimpse-folder"] is not None:
                layout["glimpse_folder"].reset(
                    path=Path(DEFAULTS["channels"][c]["glimpse-folder"])
                )
                layout["glimpse_folder"]._apply_selection()
            layout.add_child("driftlist", FileChooser(".", title="Driftlist file"))
            if DEFAULTS["channels"][c]["driftlist"] is not None:
                layout["driftlist"].reset(
                    path=Path(DEFAULTS["channels"][c]["driftlist"]).parent,
                    filename=Path(DEFAULTS["channels"][c]["driftlist"]).name,
                )
                layout["driftlist"]._apply_selection()
            layout.add_child(
                "ontarget_aoiinfo",
                FileChooser(".", title="Target molecule locations file"),
            )
            if DEFAULTS["channels"][c]["ontarget-aoiinfo"] is not None:
                layout["ontarget_aoiinfo"].reset(
                    path=Path(DEFAULTS["channels"][c]["ontarget-aoiinfo"]).parent,
                    filename=Path(DEFAULTS["channels"][c]["ontarget-aoiinfo"]).name,
                )
                layout["ontarget_aoiinfo"]._apply_selection()
            layout.add_child(
                "offtarget_aoiinfo",
                FileChooser(".", title="Off-target control locations file"),
                hide=not useOfftarget.value,
            )
            if DEFAULTS["channels"][c]["offtarget-aoiinfo"] is not None:
                layout["offtarget_aoiinfo"].reset(
                    path=Path(DEFAULTS["channels"][c]["offtarget-aoiinfo"]).parent,
                    filename=Path(DEFAULTS["channels"][c]["offtarget-aoiinfo"]).name,
                )
                layout["offtarget_aoiinfo"]._apply_selection()
            channelTabs.children = channelTabs.children + (layout,)
        channelTabs.set_title(c, f"Channel #{c}")
    channelTabs.children = channelTabs.children[:C]


def fitUI(out, DEFAULTS):
    # Fit the data
    layout = inputBox()
    layout.add_child(
        "model",
        widgets.Dropdown(
            description="Tapqir model",
            value="cosmos",
            options=avail_models,
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "channels",
        widgets.Dropdown(
            description="Channel numbers",
            value="0",
            options=[str(c) for c in range(DEFAULTS["num-channels"])],
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "cuda",
        widgets.Checkbox(
            value=DEFAULTS["cuda"],
            description="Run computations on GPU?",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "nbatch_size",
        widgets.IntText(
            value=DEFAULTS["nbatch-size"],
            description="AOI batch size",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "fbatch_size",
        widgets.IntText(
            value=DEFAULTS["fbatch-size"],
            description="Frame batch size",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "learning_rate",
        widgets.FloatText(
            value=DEFAULTS["learning-rate"],
            description="Learning rate",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "num_iter",
        widgets.IntText(
            description="Number of iterations",
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "matlab",
        widgets.Checkbox(
            value=DEFAULTS["matlab"],
            description="Save parameters in matlab format?",
            style={"description_width": "initial"},
        ),
    )
    priors = inputBox()
    for name, value in DEFAULTS["priors"].items():
        priors.add_child(
            name,
            widgets.FloatText(
                value=value, description=name, style={"description_width": "initial"}
            ),
        )
    layout.add_child("priors", widgets.Accordion(children=[priors]))
    layout["priors"].set_title(0, "Priors")
    # Fit the data
    layout.add_child("fit", widgets.Button(description="Fit the data"))
    # Callbacks
    layout["fit"].on_click(partial(fitCmd, layout=layout, out=out, DEFAULTS=DEFAULTS))
    return layout


def fitCmd(b, layout, out, DEFAULTS):
    # read priors
    DEFAULTS["priors"].update(layout["priors"].children[0].kwargs)
    with out:
        fit(
            **layout.kwargs,
            k_max=2,
            funsor=False,
            pykeops=True,
            no_input=True,
            progress_bar=tqdm_notebook,
        )

    out.clear_output(wait=True)


def showUI(out, DEFAULTS):
    # View results
    layout = inputBox()
    layout.add_child(
        "model",
        widgets.Dropdown(
            description="Tapqir model",
            value="cosmos",
            options=avail_models,
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "channels",
        widgets.Dropdown(
            description="Channel numbers",
            value="0",
            options=[str(c) for c in range(DEFAULTS["num-channels"])],
            style={"description_width": "initial"},
        ),
    )
    layout.add_child(
        "show_fov",
        widgets.Checkbox(
            value=True,
            description="Show FOV images",
            indent=False,
        ),
    )
    # Load results
    layout.add_child(
        "show",
        widgets.Button(
            description="Load results",
            layout=widgets.Layout(margin="10px 0px 10px 0px"),
        ),
    )
    controls = widgets.HBox()
    view = widgets.Output()
    fov_controls = inputBox()
    params = widgets.VBox(children=[controls, view, fov_controls])
    summary = widgets.Output()
    sum_view = widgets.Accordion(children=[params, summary])
    sum_view.set_title(0, "AOI images and parameters")
    sum_view.set_title(1, "Summary statistics")
    sum_view.selected_index = 0
    layout.add_child("view", sum_view)
    # Callbacks
    layout["show"].on_click(partial(showCmd, layout=layout, out=out))
    return layout


def showCmd(b, layout, out):
    layout.toggle_hide(names=("show",))
    params, summary = layout["view"].children
    controls, view, fov_controls = params.children
    with out:
        logger.info("Loading results ...")
        show_ret = show(
            **layout.kwargs,
            n=0,
            f1=None,
            f2=None,
            gui=True,
        )
        if show_ret == 1:
            out.clear_output(wait=True)
            return
        model, fig, item, ax, fov = show_ret
    with view:
        plt.show()
    with summary:
        display(model.summary)
    n = widgets.BoundedIntText(
        value=0,
        min=0,
        max=model.data.Nt - 1,
        description=f"AOI (0-{model.data.Nt-1})",
        style={"description_width": "initial"},
        layout={"width": "150px"},
    )
    n_counter = widgets.BoundedIntText(
        min=0,
        max=model.data.Nt - 1,
    )
    f1_text = widgets.BoundedIntText(
        value=0,
        min=0,
        max=model.data.F - 15,
        step=1,
        description=f"Frame (0-{model.data.F-1})",
        style={"description_width": "initial"},
        layout={"width": "300px"},
    )
    f1_counter = widgets.BoundedIntText(
        min=0,
        max=model.data.F - 15,
    )
    f1_slider = widgets.IntSlider(
        value=0,
        min=0,
        max=model.data.F - 15,
        step=1,
        continuous_update=True,
        readout=False,
        readout_format="d",
        layout={"width": "210px"},
    )
    f1_box = widgets.VBox(layout=widgets.Layout(width="310px"))
    f1_incr = widgets.Button(description="+15", layout=widgets.Layout(width="45px"))
    f1_decr = widgets.Button(description="-15", layout=widgets.Layout(width="45px"))
    f1_controls = widgets.HBox(
        children=[f1_decr, f1_slider, f1_incr],
        layout=widgets.Layout(width="305px"),
    )
    f1_box.children = [f1_text, f1_controls]
    widgets.jslink((f1_slider, "value"), (f1_text, "value"))
    widgets.jslink((f1_text, "value"), (f1_counter, "value"))
    widgets.jslink((n, "value"), (n_counter, "value"))
    zoom = widgets.Checkbox(
        value=False,
        description="Zoom out frames ['z']",
        indent=False,
        layout={"width": "240px"},
    )
    labels = widgets.Checkbox(
        value=False,
        description="Show labels",
        indent=False,
        layout={"width": "240px"},
    )
    targets = widgets.Checkbox(
        value=False,
        description="Show target location ['o']",
        indent=False,
        layout={"width": "240px"},
    )
    nonspecific = widgets.Checkbox(
        value=True,
        description="Show non-specific spots ['n']",
        indent=False,
        layout={"width": "240px"},
    )
    exclude_aoi = widgets.Checkbox(
        value=not model.data.mask[0],
        description="Exclude AOI from analysis ['e']",
        indent=False,
        layout={"width": "240px"},
    )
    checkboxes = widgets.VBox(layout=widgets.Layout(width="250px"))
    if model.data.labels is None:
        checkboxes.children = [zoom, targets, nonspecific, exclude_aoi]
    else:
        checkboxes.children = [zoom, targets, nonspecific, exclude_aoi, labels]
    controls.children = [n, f1_box, checkboxes]
    # fov controls
    if fov is not None:
        for dtype in fov.dtypes:
            fov_controls.add_child(
                dtype, widgets.Checkbox(value=True, description=f"Show {dtype} AOIs")
            )
            fov_controls[dtype].observe(
                partial(
                    showAOIs,
                    fov=fov,
                    n=n_counter,
                    item=item,
                    fig=fig,
                ),
                names="value",
            )
    fov_controls.add_child("save_data", widgets.Button(description="Save data"))
    # callbacks
    n.observe(
        partial(
            updateParams,
            f1=f1_slider,
            model=model,
            fig=fig,
            item=item,
            ax=ax,
            targets=targets,
            nonspecific=nonspecific,
            labels=labels,
            fov_controls=fov_controls,
            exclude_aoi=exclude_aoi,
        ),
        names="value",
    )
    f1_slider.observe(
        partial(
            updateRange,
            n=n,
            model=model,
            fig=fig,
            item=item,
            ax=ax,
            zoom=zoom,
            targets=targets,
            fov=fov,
        ),
        names="value",
    )
    f1_incr.on_click(partial(incrementRange, x=15, counter=f1_counter))
    f1_decr.on_click(partial(incrementRange, x=-15, counter=f1_counter))
    zoom.observe(
        partial(zoomOut, f1=f1_slider, model=model, fig=fig, item=item, ax=ax),
        names="value",
    )
    labels.observe(
        partial(showLabels, n=n_counter, model=model, item=item, ax=ax),
        names="value",
    )
    targets.observe(
        partial(showTargets, n=n_counter, f1=f1_slider, model=model, item=item, ax=ax),
        names="value",
    )
    nonspecific.observe(
        partial(showNonspecific, n=n_counter, model=model, item=item, ax=ax),
        names="value",
    )
    exclude_aoi.observe(
        partial(excludeAOI, n=n_counter, model=model, item=item),
        names="value",
    )
    fov_controls["save_data"].on_click(
        partial(saveData, data=model.data, path=model.path, out=out)
    )
    # mouse click UI
    fig.canvas.mpl_connect(
        "button_release_event",
        partial(onFrameClick, counter=f1_counter),
    )
    # key press UI
    d = Event(source=layout, watched_events=["keyup"], prevent_default_action=True)
    d.on_dom_event(
        partial(
            onKeyPress,
            n=n_counter,
            f1=f1_counter,
            zoom=zoom,
            targets=targets,
            nonspecific=nonspecific,
            exclude_aoi=exclude_aoi,
        )
    )
    with out:
        logger.info("Loading results: Done")

    out.clear_output(wait=True)
    b.disabled = True


def saveData(b, data, path, out):
    with out:
        logger.info("Saving data ...")
        save(data, path)


def showAOIs(checked, fov, n, item, fig):
    dtype = "ontarget" if "ontarget" in checked.owner.description else "offtarget"
    n = get_value(n)
    for key, value in item.items():
        if key.startswith("aoi_"):
            i = int(key.split("_")[1])
            if n == i:
                continue
            if dtype == "ontarget" and i >= fov.N:
                continue
            if dtype == "offtarget" and i < fov.N:
                continue
            item[key].set(visible=get_value(checked))
    fig.canvas.draw()


def onKeyPress(event, n, f1, zoom, targets, nonspecific, exclude_aoi):
    key = event["key"]
    if key == "h" or key == "ArrowLeft":
        f1.value = f1.value - 15
    elif key == "l" or key == "ArrowRight":
        f1.value = f1.value + 15
    elif key == "j" or key == "ArrowDown":
        n.value = n.value - 1
    elif key == "k" or key == "ArrowUp":
        n.value = n.value + 1
    elif key == "z":
        zoom.value = not zoom.value
    elif key == "o":
        targets.value = not targets.value
    elif key == "n":
        nonspecific.value = not nonspecific.value
    elif key == "e":
        exclude_aoi.value = not exclude_aoi.value


def onFrameClick(event, counter):
    counter.value = int(event.xdata)


def incrementRange(name, x, counter):
    counter.value = counter.value + x


def updateParams(
    n, f1, model, fig, item, ax, targets, nonspecific, labels, fov_controls, exclude_aoi
):
    n_old = n.old
    n = get_value(n)
    f1 = get_value(f1)
    f2 = f1 + 15
    c = model.cdx
    color = "C2" if model.data.mask[n] else "C7"

    exclude_aoi.value = not model.data.mask[n]

    frames = torch.arange(f1, f2)
    img_ideal = (
        model.data.offset.mean
        + model.params["background"]["Mean"][n, frames, None, None]
    )
    gaussian = gaussian_spots(
        model.params["height"]["Mean"][:, n, frames],
        model.params["width"]["Mean"][:, n, frames],
        model.params["x"]["Mean"][:, n, frames],
        model.params["y"]["Mean"][:, n, frames],
        model.data.xy[n, frames, c],
        model.data.P,
    )
    img_ideal = img_ideal + gaussian.sum(-4)
    for i, f in enumerate(range(f1, f2)):
        ax[f"image_{i}"].set_title(rf"${f}$", fontsize=9)
        item[f"image_{i}"].set_data(model.data.images[n, f, c].numpy())
        item[f"ideal_{i}"].set_data(img_ideal[i].numpy())
        if targets.value:
            item[f"target_{i}"].remove()
            item[f"target_{i}"] = ax[f"image_{i}"].scatter(
                model.data.x[n, f, c].item(),
                model.data.y[n, f, c].item(),
                c="C0",
                s=40,
                marker="+",
            )

    params = ["p_specific", "z_map", "height", "width", "x", "y", "background", "chi2"]
    if labels.value:
        params += ["labels"]
    theta_mask = model.params["theta_probs"][:, n] > 0.5
    j_mask = (model.params["m_probs"][:, n] > 0.5) & ~theta_mask
    for p in params:
        if p == "p_specific":
            item[p].set_ydata(model.params[p][n])
            item[p].set_color(color)
        elif p in {"z_map"}.intersection(model.params.keys()):
            item[p].set_ydata(model.params[p][n])
        elif p == "chi2":
            item[p].set_ydata(model.params[p]["values"][n])
        elif p == "background":
            item[f"{p}_fill"].remove()
            item[f"{p}_fill"] = ax[p].fill_between(
                torch.arange(0, model.data.F),
                model.params[p]["LL"][n],
                model.params[p]["UL"][n],
                alpha=0.3,
                color="k",
            )
            item[f"{p}_mean"].set_ydata(model.params[p]["Mean"][n])
        elif p == "labels":
            item[p].set_ydata(model.data.labels["z"][n, :, c])
        elif p in ["height", "width", "x", "y"]:
            # target-nonspecific spots
            if nonspecific.value:
                for k in range(model.K):
                    f_mask = j_mask[k]
                    mean = model.params[p]["Mean"][k, n] * f_mask
                    ll = model.params[p]["LL"][k, n] * f_mask
                    ul = model.params[p]["UL"][k, n] * f_mask
                    item[f"{p}_nonspecific{k}_mean"].remove()
                    (item[f"{p}_nonspecific{k}_mean"],) = ax[p].plot(
                        torch.arange(0, model.data.F)[f_mask],
                        mean[f_mask],
                        "o",
                        ms=2,
                        lw=1,
                        color=f"C{k}",
                    )
                    item[f"{p}_nonspecific{k}_fill"].remove()
                    item[f"{p}_nonspecific{k}_fill"] = ax[p].fill_between(
                        torch.arange(0, model.data.F),
                        ll,
                        ul,
                        where=f_mask,
                        alpha=0.3,
                        color=f"C{k}",
                    )
            # target-specific spots
            f_mask = theta_mask.sum(0).bool()
            mean = (model.params[p]["Mean"][:, n] * theta_mask).sum(0)
            ll = (model.params[p]["LL"][:, n] * theta_mask).sum(0)
            ul = (model.params[p]["UL"][:, n] * theta_mask).sum(0)
            item[f"{p}_specific_fill"].remove()
            item[f"{p}_specific_fill"] = ax[p].fill_between(
                torch.arange(0, model.data.F),
                ll,
                ul,
                where=f_mask,
                alpha=0.3,
                color=color,
            )
            item[f"{p}_specific_mean"].remove()
            (item[f"{p}_specific_mean"],) = ax[p].plot(
                torch.arange(0, model.data.F)
                if p == "height"
                else torch.arange(0, model.data.F)[f_mask],
                mean if p == "height" else mean[f_mask],
                "-" if p == "height" else "o",
                ms=2,
                color=color,
            )

    ax["glimpse"].set_title(rf"AOI ${n}$, Frame ${f1}$", fontsize=9)
    item[f"aoi_{n}"].set_edgecolor(color)
    item[f"aoi_{n}"].set(zorder=2, visible=True)
    n_old_dtype = "ontarget" if n_old < model.data.N else "offtarget"
    n_old_visible = fov_controls[n_old_dtype].value
    colors = {"ontarget": "#AA3377", "offtarget": "#CCBB44"}
    item[f"aoi_{n_old}"].set_edgecolor(colors[n_old_dtype])
    item[f"aoi_{n_old}"].set(zorder=1, visible=n_old_visible)
    fig.canvas.draw()


def updateRange(f1, n, model, fig, item, ax, zoom, targets, fov):
    n = get_value(n)
    f1 = get_value(f1)
    f2 = f1 + 15
    c = model.cdx

    frames = torch.arange(f1, f2)
    img_ideal = (
        model.data.offset.mean
        + model.params["background"]["Mean"][n, frames, None, None]
    )
    gaussian = gaussian_spots(
        model.params["height"]["Mean"][:, n, frames],
        model.params["width"]["Mean"][:, n, frames],
        model.params["x"]["Mean"][:, n, frames],
        model.params["y"]["Mean"][:, n, frames],
        model.data.xy[n, frames, c],
        model.data.P,
    )
    img_ideal = img_ideal + gaussian.sum(-4)
    for i, f in enumerate(range(f1, f2)):
        ax[f"image_{i}"].set_title(rf"${f}$", fontsize=9)
        item[f"image_{i}"].set_data(model.data.images[n, f, c].numpy())
        item[f"ideal_{i}"].set_data(img_ideal[i].numpy())
        if targets.value:
            item[f"target_{i}"].remove()
            item[f"target_{i}"] = ax[f"image_{i}"].scatter(
                model.data.x[n, f, c].item(),
                model.data.y[n, f, c].item(),
                c="C0",
                s=40,
                marker="+",
            )

    if not zoom.value:
        for key, a in ax.items():
            if key.startswith("image") or key.startswith("ideal") or key == "glimpse":
                continue
            a.set_xlim(f1 - 0.5, f2 - 0.5)
    else:
        for p in [
            "z_map",
            "p_specific",
            "height",
            "width",
            "x",
            "y",
            "background",
            "chi2",
        ]:
            item[f"{p}_vspan"].remove()
            item[f"{p}_vspan"] = ax[p].axvspan(f1, f2, facecolor="C0", alpha=0.3)
    if fov is not None:
        fov.plot(
            fov.dtypes,
            model.data.P,
            n=n,
            f=f1,
            save=False,
            ax=ax["glimpse"],
            item=item,
        )
    fig.canvas.draw()


def zoomOut(checked, f1, model, fig, item, ax):
    checked = get_value(checked)
    f1_value = get_value(f1)
    if checked:
        f1 = 0
        f2 = model.data.F
        for p in [
            "z_map",
            "p_specific",
            "height",
            "width",
            "x",
            "y",
            "background",
            "chi2",
        ]:
            item[f"{p}_vspan"] = ax[p].axvspan(
                f1_value, f1_value + 15, facecolor="C0", alpha=0.3
            )
    else:
        f1 = f1_value
        f2 = f1 + 15
        for p in [
            "z_map",
            "p_specific",
            "height",
            "width",
            "x",
            "y",
            "background",
            "chi2",
        ]:
            item[f"{p}_vspan"].remove()
    for key, a in ax.items():
        if key.startswith("image") or key.startswith("ideal") or key == "glimpse":
            continue
        a.set_xlim(f1 - 0.5, f2 - 0.5)
    fig.canvas.draw()


def showLabels(checked, n, model, item, ax):
    checked = get_value(checked)
    n = get_value(n)
    if checked:
        (item["labels"],) = ax["z_map"].plot(
            torch.arange(0, model.data.F),
            model.data.labels["z"][n, :, model.cdx],
            "-",
            lw=1,
            color="r",
        )
    else:
        item["labels"].remove()


def showTargets(checked, n, f1, model, item, ax):
    checked = get_value(checked)
    n = get_value(n)
    f1 = get_value(f1)
    f2 = f1 + 15
    c = model.cdx
    if checked:
        for i, f in enumerate(range(f1, f2)):
            item[f"target_{i}"] = ax[f"image_{i}"].scatter(
                model.data.x[n, f, c].item(),
                model.data.y[n, f, c].item(),
                c="C0",
                s=40,
                marker="+",
            )
    else:
        for i, f in enumerate(range(f1, f2)):
            item[f"target_{i}"].remove()


def showNonspecific(checked, n, model, item, ax):
    checked = get_value(checked)
    n = get_value(n)
    for p in ["height", "width", "x", "y"]:
        if checked:
            theta_mask = model.params["theta_probs"][:, n] > 0.5
            j_mask = (model.params["m_probs"][:, n] > 0.5) & ~theta_mask
            # target-nonspecific spots
            for k in range(model.K):
                f_mask = j_mask[k]
                mean = model.params[p]["Mean"][k, n] * f_mask
                ll = model.params[p]["LL"][k, n] * f_mask
                ul = model.params[p]["UL"][k, n] * f_mask
                (item[f"{p}_nonspecific{k}_mean"],) = ax[p].plot(
                    torch.arange(0, model.data.F)[f_mask],
                    mean[f_mask],
                    "o",
                    ms=2,
                    lw=1,
                    color=f"C{k}",
                )
                item[f"{p}_nonspecific{k}_fill"] = ax[p].fill_between(
                    torch.arange(0, model.data.F),
                    ll,
                    ul,
                    where=f_mask,
                    alpha=0.3,
                    color=f"C{k}",
                )
        else:
            for k in range(model.K):
                item[f"{p}_nonspecific{k}_fill"].remove()
                item[f"{p}_nonspecific{k}_mean"].remove()


def excludeAOI(checked, n, model, item):
    checked = get_value(checked)
    n = get_value(n)
    model.data.mask[n] = not checked
    color = "C7" if checked else "C2"
    item["p_specific"].set_color(color)
    item["height_specific_mean"].set_color(color)
    item["height_specific_fill"].set_color(color)
    item["width_specific_mean"].set_color(color)
    item["width_specific_fill"].set_color(color)
    item["x_specific_mean"].set_color(color)
    item["x_specific_fill"].set_color(color)
    item["y_specific_mean"].set_color(color)
    item["y_specific_fill"].set_color(color)
    item[f"aoi_{n}"].set_edgecolor(color)


def logUI(out):
    # View logs
    layout = widgets.VBox()
    # Load logs
    showLog = widgets.Button(description="(Re)-load logs")
    logView = widgets.Output(layout={"max_height": "850px", "overflow": "auto"})
    # Layout
    layout.children = [
        showLog,
        logView,
    ]
    # Callbacks
    showLog.on_click(partial(logCmd, layout=layout, logView=logView, out=out))
    return layout


def logCmd(b, layout, logView, out):
    with out:
        logger.info("Loading logs ...")
    with logView:
        log()
    with out:
        logger.info("Loading logs: Done")
    logView.clear_output(wait=True)
    out.clear_output(wait=True)


def app():
    import pkg_resources

    nbpath = pkg_resources.resource_filename("tapqir", "app.ipynb")
    os.system("voila " + nbpath)


def run():
    from tapqir.main import DEFAULTS

    display(initUI(DEFAULTS))
